import typing as t

if t.TYPE_CHECKING:
    from flama import types, websockets

__all__ = ["EndpointProtocol", "HTTPEndpointProtocol", "WebSocketEndpointProtocol"]


class EndpointProtocol(t.Protocol):
    def __init__(self, scope: "types.Scope", receive: "types.Receive", send: "types.Send") -> None: ...

    def __await__(self) -> t.Generator: ...

    @classmethod
    def allowed_handlers(cls) -> dict[str, t.Callable]: ...

    async def dispatch(self) -> None: ...


class HTTPEndpointProtocol(EndpointProtocol, t.Protocol):
    @classmethod
    def allowed_methods(cls) -> set[str]: ...

    @property
    def handler(self) -> t.Callable: ...


class WebSocketEndpointProtocol(EndpointProtocol, t.Protocol):
    encoding: t.Union["types.Encoding", None] = None

    async def on_connect(self, websocket: "websockets.WebSocket") -> None: ...

    async def on_receive(self, websocket: "websockets.WebSocket", data: "types.Data") -> None: ...

    async def on_disconnect(self, websocket: "websockets.WebSocket", websocket_code: "types.Code") -> None: ...



import pytest
from _pytest.mark import param

from flama import endpoints, types, websockets


class TestCaseDataValidation:
    @pytest.mark.parametrize(
        "request_params,response_status,response_json",
        [
            # JSON
            param({"json": {"abc": 123}}, 200, {"data": {"abc": 123}}, id="valid json body"),
            param({}, 200, {"data": None}, id="empty json body"),
            # Urlencoding
            param({"data": {"abc": 123}}, 200, {"data": {"abc": "123"}}, id="valid urlencoded body"),
            param(
                {"headers": {"content-type": "application/x-www-form-urlencoded"}},
                200,
                {"data": None},
                id="empty urlencoded body",
            ),
            # Multipart
            param(
                {"files": {"a": ("b", b"123")}, "data": {"b": "42"}},
                200,
                {"data": {"a": {"filename": "b", "content": "123"}, "b": "42"}},
                id="multipart",
            ),
            # Misc
            param({"content": b"...", "headers": {"content-type": "unknown"}}, 415, None, id="unknown body type"),
            param(
                {"content": b"...", "headers": {"content-type": "application/json"}}, 400, None, id="json parse failure"
            ),
        ],
    )
    async def test_request_data(self, request_params, response_status, response_json, app, client):
        @app.route("/request_data/", methods=["POST"])
        async def get_request_data(data: types.RequestData):
            try:
                data = types.RequestData(
                    {
                        key: value
                        if not hasattr(value, "filename")
                        else {"filename": value.filename, "content": (await value.read()).decode("utf-8")}
                        for key, value in data.items()
                    }
                )
            except Exception:
                pass

            return {"data": data}

        response = await client.request("post", "/request_data/", **request_params)
        assert response.status_code == response_status, str(response.content)
        if response_json is not None:
            assert response.json() == response_json

    @pytest.mark.skip(reason="Cannot test websockets with current client")  # CAVEAT: Client doesn't support websockets
    @pytest.mark.parametrize(
        "encoding,send_method,data,expected_result",
        [
            # bytes
            param("bytes", "bytes", b"foo", {"type": "websocket.send", "bytes": b"foo"}, id="bytes"),
            param(
                "bytes",
                "text",
                b"foo",
                {"type": "websocket.close", "code": 1003, "reason": ""},
                id="bytes wrong format",
            ),
            # text
            param("text", "text", "foo", {"type": "websocket.send", "text": "foo"}, id="text"),
            param(
                "text", "bytes", "foo", {"type": "websocket.close", "code": 1003, "reason": ""}, id="text wrong format"
            ),
            # json
            param(
                "json",
                "json",
                {"foo": "bar"},
                {"type": "websocket.send", "text": '{"foo":"bar"}'},
                id="json from json",
            ),
            param(
                "json",
                "text",
                '{"foo": "bar"}',
                {"type": "websocket.send", "text": '{"foo":"bar"}'},
                id="json from text",
            ),
            param(
                "json",
                "bytes",
                b'{"foo": "bar"}',
                {"type": "websocket.send", "text": '{"foo":"bar"}'},
                id="json from bytes",
            ),
            param(
                "json",
                "bytes",
                b'{"foo":',
                {"type": "websocket.close", "code": 1003, "reason": ""},
                id="json wrong format",
            ),
        ],
    )
    def test_websocket_message_data(self, encoding, send_method, data, expected_result, app, client):
        encoding_ = encoding

        @app.websocket_route("/websocket/")
        class Endpoint(endpoints.WebSocketEndpoint):
            encoding = encoding_

            async def on_receive(self, websocket: websockets.WebSocket, data: types.Data):
                await getattr(websocket, f"send_{encoding}")(data)

        with client.websocket_connect("/websocket/") as ws:
            getattr(ws, f"send_{send_method}")(data)
            message = ws.receive()

        assert message == expected_result



import typing as t

if t.TYPE_CHECKING:
    from flama.injection.components import Component
    from flama.injection.resolver import Parameter

__all__ = ["ComponentError", "ComponentNotFound"]


class ComponentError(Exception): ...


class ComponentNotFound(ComponentError):
    def __init__(
        self,
        parameter: "Parameter",
        component: t.Optional["Component"] = None,
        function: t.Optional[t.Callable] = None,
    ):
        self.parameter = parameter
        self.component = component
        self.function = function

    def __str__(self) -> str:
        msg = f"No component able to handle parameter '{self.parameter.name}'"
        if self.component:
            msg += f" in component '{self.component.__str__()}'"
        if self.function:
            msg += f" for function '{self.function.__name__}'"

        return msg



import tempfile
import typing as t

import pytest


class ConfigFileFactory:
    def __init__(self):
        self._factories = {
            "config_no_sections": self._config_no_sections,
            "config": self._config,
            "json": self._json,
            "yaml": self._yaml,
            "toml": self._toml,
        }

        self._file: tempfile._TemporaryFileWrapper

    @property
    def file(self) -> tempfile._TemporaryFileWrapper:
        if not getattr(self, "_file"):
            raise AttributeError("Config file not initialized")

        return self._file

    @file.setter
    def file(self, f: tempfile._TemporaryFileWrapper) -> None:
        self._file = f

    @file.deleter
    def file(self) -> None:
        del self._file

    def __enter__(self) -> "ConfigFileFactory":
        self.file = tempfile.NamedTemporaryFile("w+")
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        if self.file:
            self.file.__exit__(exc_type, exc_value, traceback)
            del self.file

    def generate(self, factory: str) -> dict[str, t.Any]:
        try:
            result = self._factories[factory]()
            self.file.seek(0)
            return result
        except KeyError:
            raise ValueError(f"Config file factory '{factory}' not found")

    def _config_no_sections(self) -> dict[str, t.Any]:
        self.file.write("foo = 1")
        return {"foo": "1"}

    def _config(self) -> dict[str, t.Any]:
        self.file.write("[foo]\nbar = 1")
        return {"foo": {"bar": "1"}}

    def _json(self) -> dict[str, t.Any]:
        self.file.write('{"foo": 1}')
        return {"foo": 1}

    def _yaml(self) -> dict[str, t.Any]:
        self.file.write("foo:\n    bar: 1")
        return {"foo": {"bar": 1}}

    def _toml(self) -> dict[str, t.Any]:
        self.file.write("[foo]\nbar = 1")
        return {"foo": {"bar": 1}}


@pytest.fixture(scope="function")
def config_file(request):
    with ConfigFileFactory() as factory:
        config_obj = factory.generate(request.param)
        yield (factory.file.name, config_obj)



import dataclasses
import typing as t

from flama.ddd.workers.sqlalchemy import SQLAlchemyWorker
from flama.exceptions import ApplicationError

if t.TYPE_CHECKING:
    from flama import Flama
    from flama.ddd.repositories.sqlalchemy import SQLAlchemyTableRepository


@dataclasses.dataclass
class Repositories:
    registered: dict[str, type["SQLAlchemyTableRepository"]] = dataclasses.field(default_factory=dict)
    initialised: t.Optional[dict[str, "SQLAlchemyTableRepository"]] = None

    def init(self, *args: t.Any, **kwargs: t.Any) -> None:
        self.initialised = {r: cls(*args, **kwargs) for r, cls in self.registered.items()}

    def delete(self) -> None:
        self.initialised = None


class FlamaWorker(SQLAlchemyWorker):
    """The worker used by Flama Resources."""

    def __init__(self, app: t.Optional["Flama"] = None):
        """Initialize the worker.

        This special worker is used to handle the repositories created by Flama Resources.

        :param app: The application instance.
        """

        super().__init__(app)
        self._resources_repositories = Repositories()

    def add_repository(self, name: str, repository: type["SQLAlchemyTableRepository"]) -> None:
        """Register a repository.

        :param name: The name of the repository.
        :param repository: The repository class.
        """
        self._resources_repositories.registered[name] = repository

    def remove_repository(self, name: str) -> None:
        """Deregister a repository.

        :param name: The name of the repository.
        """
        del self._resources_repositories.registered[name]

    @property
    def repositories(self) -> dict[str, "SQLAlchemyTableRepository"]:
        """Get the initialized repositories.

        :retirns: The initialized repositories.
        :raises ApplicationError: If the repositories are not initialized.
        """
        if not self._resources_repositories.initialised:
            raise ApplicationError("Repositories not initialized")

        return self._resources_repositories.initialised

    async def begin(self) -> None:
        """Start a unit of work.

        Initialize the connection, begin a transaction, and create the repositories.
        """
        await self.set_up()
        args, kwargs = await self.repository_params()
        self._resources_repositories.init(*args, **kwargs)

    async def end(self, *, rollback: bool = False) -> None:
        """End a unit of work.

        Close the connection, commit or rollback the transaction, and delete the repositories.

        :param rollback: If the unit of work should be rolled back.
        """
        await self.tear_down(rollback=rollback)
        self._resources_repositories.delete()



import abc
import hmac

__all__ = ["SignAlgorithm", "HMACAlgorithm"]


class SignAlgorithm(abc.ABC):
    """Abstract class for signature algorithms."""

    @abc.abstractmethod
    def sign(self, message: bytes, key: bytes) -> bytes:
        """Sign a message using the given key.

        :param message: Message to sign.
        :param key: Key used to sign the message.
        :return: Signature.
        """
        ...

    @abc.abstractmethod
    def verify(self, message: bytes, signature: bytes, key) -> bool:
        """Verify the signature of a message.

        :param message: Message to verify.
        :param signature: Signed message.
        :param key: Key used to sign the message.
        :return: True if the signature is valid, False otherwise.
        """
        ...


class HMACAlgorithm(SignAlgorithm):
    """HMAC using SHA algorithms for JWS."""

    def __init__(self, sha):
        self.hash_algorithm = sha

    def sign(self, message: bytes, key: bytes) -> bytes:
        """Sign a message using the given key.

        :param message: Message to sign.
        :param key: Key used to sign the message.
        :return: Signature.
        """
        return hmac.new(key, message, self.hash_algorithm).digest()

    def verify(self, message: bytes, signature: bytes, key) -> bool:
        """Verify the signature of a message.

        :param message: Message to verify.
        :param signature: Signed message.
        :param key: Key used to sign the message.
        :return: True if the signature is valid, False otherwise.
        """
        return hmac.compare_digest(signature, hmac.new(key, message, self.hash_algorithm).digest())



import functools
import json
import typing as t

import click

from flama.http import EnhancedJSONEncoder
from flama.models import ModelComponentBuilder

if t.TYPE_CHECKING:
    from flama.models import ModelComponent

__all__ = ["model", "command"]


@click.group(name="model")
@click.argument("flama-model-path", envvar="FLAMA_MODEL_PATH")
@click.pass_context
def command(ctx: click.Context, flama_model_path: str):
    """Interact with an ML model without server.

    This command is used to directly interact with an ML model without the need of a server. This command can be used
    to perform any operation that is supported by the model, such as inspect, or predict.
    <FLAMA_MODEL_PATH> is the path of the model to be used, e.g. 'path/to/model.flm'. This can be passed
    directly as argument of the command line, or by environment variable.
    """
    try:
        ctx.obj = ModelComponentBuilder.load(flama_model_path)
    except FileNotFoundError:
        raise click.BadParameter("Model file not found.")


@command.command(name="inspect", context_settings={"auto_envvar_prefix": "FLAMA"})
@click.option("-p", "--pretty", is_flag=True, default=False, help="Pretty print the model inspection.")
@click.pass_obj
def inspect(model: "ModelComponent", pretty: bool):
    """Inspect an ML model.

    This command is used to inspect an ML model without the need of a server. This command can be used to extract the
    ML model metadata, including the ID, time when the model was created, information of the
    framework, and the model info; and the list of artifacts packaged with the model.
    """
    dump_func = functools.partial(json.dumps, cls=EnhancedJSONEncoder)
    if pretty:
        dump_func = functools.partial(dump_func, sort_keys=True, indent=4)

    click.echo(dump_func(model.model.inspect()))


@command.command(name="predict", context_settings={"auto_envvar_prefix": "FLAMA"})
@click.option(
    "-f",
    "--file",
    "input_file",
    type=click.File("r"),
    default="-",
    help="File to be used as input for the model prediction in JSON format. (default: stdin).",
)
@click.option(
    "-o",
    "--output",
    "output_file",
    type=click.File("w"),
    default="-",
    help="File to be used as output for the model prediction in JSON format. (default: stdout).",
)
@click.option("-p", "--pretty", is_flag=True, default=False, help="Pretty print the model prediction.")
@click.pass_obj
def predict(model: "ModelComponent", input_file, output_file, pretty: bool):
    """Make a prediction using an ML model.

    This command is used to make a prediction using an ML model without the need of a server. It can be used for
    batch predictions, so both input and output arguments must be json files containing a list of input values, each
    input value being a list of values associated to the input of the model. The output will be the list of predictions
    associated to the input, with each prediction being a list of values representing the output of the model.

    Example:

    - input.json:
    [[0, 0], [0, 1], [1, 0], [1, 1]]

    - output.json:
    [[0], [1], [1], [0]]
    """
    try:
        data = json.load(input_file)
    except json.JSONDecodeError:
        raise click.BadParameter("Input file must be a valid json file.")

    dump_func = functools.partial(json.dumps, cls=EnhancedJSONEncoder)
    if pretty:
        dump_func = functools.partial(dump_func, sort_keys=True, indent=4)

    click.echo(dump_func(model.model.predict(data)), output_file)


model: t.Callable = command.callback  # type: ignore[assignment]



from unittest.mock import AsyncMock, MagicMock, call, patch

import pytest
from sqlalchemy.ext.asyncio import AsyncConnection

from flama.ddd.repositories.sqlalchemy import SQLAlchemyRepository
from flama.ddd.workers.sqlalchemy import SQLAlchemyWorker


class TestCaseSQLAlchemyWorker:
    @pytest.fixture(scope="function")
    def worker(self, client):
        class FooWorker(SQLAlchemyWorker):
            bar: SQLAlchemyRepository

        return FooWorker(client.app)

    def test_init(self, app):
        worker = SQLAlchemyWorker(app)

        assert worker._app == app
        assert not hasattr(worker, "_connection")

    def test_connection(self, worker):
        with pytest.raises(AttributeError, match="Connection not initialized"):
            worker.connection

    def test_transaction(self, worker):
        with pytest.raises(AttributeError, match="Transaction not started"):
            worker.transaction

    async def test_set_up(self, app, worker):
        connection_mock = AsyncMock()
        transaction_mock = AsyncMock()

        with patch.multiple(
            app.sqlalchemy,
            open_connection=AsyncMock(return_value=connection_mock),
            begin_transaction=AsyncMock(return_value=transaction_mock),
        ):
            await worker.set_up()

            assert worker._connection == connection_mock
            assert worker._transaction == transaction_mock
            assert app.sqlalchemy.open_connection.await_args_list == [call()]
            assert app.sqlalchemy.begin_transaction.await_args_list == [call(connection_mock)]

    @pytest.mark.parametrize(
        ["rollback"],
        (
            pytest.param(True, id="rollback"),
            pytest.param(False, id="commit"),
        ),
    )
    async def test_tear_down(self, app, worker, rollback):
        worker._connection = connection_mock = AsyncMock()
        worker._transaction = transaction_mock = AsyncMock()

        with patch.multiple(app.sqlalchemy, end_transaction=AsyncMock(), close_connection=AsyncMock()):
            await worker.tear_down(rollback=rollback)

            assert not hasattr(worker, "_transaction")
            assert not hasattr(worker, "_connection")
            assert app.sqlalchemy.end_transaction.await_args_list == [call(transaction_mock, rollback=rollback)]
            assert app.sqlalchemy.close_connection.await_args_list == [call(connection_mock)]

    async def test_begin(self, worker):
        worker._connection = AsyncMock()

        with patch.object(worker, "set_up"):
            assert not hasattr(worker, "bar")

            await worker.begin()

            assert worker.set_up.await_args_list == [call()]
            assert hasattr(worker, "bar")
            assert isinstance(worker.bar, SQLAlchemyRepository)

    @pytest.mark.parametrize(
        ["rollback"],
        (
            pytest.param(True, id="rollback"),
            pytest.param(False, id="commit"),
        ),
    )
    async def test_end(self, worker, rollback):
        worker.bar = MagicMock()

        with patch.object(worker, "tear_down"):
            assert hasattr(worker, "bar")

            await worker.end(rollback=rollback)

            assert worker.tear_down.await_args_list == [call(rollback=rollback)]
            assert not hasattr(worker, "bar")

    async def test_commit(self, worker):
        commit_mock = AsyncMock()
        connection_mock = AsyncMock(spec=AsyncConnection)
        connection_mock.commit = commit_mock

        with patch.object(worker, "_connection", new=connection_mock, create=True):
            await worker.commit()
            assert commit_mock.await_args_list == [call()]

    async def test_rollback(self, worker):
        rollback_mock = AsyncMock()
        connection_mock = AsyncMock(spec=AsyncConnection)
        connection_mock.rollback = rollback_mock

        with patch.object(worker, "_connection", new=connection_mock, create=True):
            await worker.rollback()
            assert rollback_mock.await_args_list == [call()]



import logging
import typing as t

from flama import concurrency, exceptions, types, url
from flama.injection import Component, Components
from flama.routing.routes.base import BaseRoute

if t.TYPE_CHECKING:
    from flama.applications import Flama

__all__ = ["Mount"]

logger = logging.getLogger(__name__)


class Mount(BaseRoute):
    def __init__(
        self,
        path: str,
        app: t.Optional[types.App] = None,
        *,
        routes: t.Optional[t.Sequence[BaseRoute]] = None,
        components: t.Optional[t.Sequence[Component]] = None,
        name: t.Optional[str] = None,
        tags: t.Optional[dict[str, t.Any]] = None,
    ):
        """A mount point for adding a nested ASGI application or a list of routes.

        :param path: URL path.
        :param app: ASGI application.
        :param routes: List of routes.
        :param components: Components registered under this mount point.
        :param name: Mount name.
        :param tags: Mount tags.
        """
        if app is None and routes is None:
            raise exceptions.ApplicationError("Either 'path' and 'app' or 'mount' variables are needed")

        if app is None:
            from flama.routing.router import Router

            app = Router(routes=routes, components=components)

        super().__init__(path, app, name=name, tags=tags)

    async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        if scope["type"] in ("http", "websocket") or (
            scope["type"] == "lifespan" and types.is_flama_instance(self.app)
        ):
            await self.handle(types.Scope({**scope, **self.route_scope(scope)}), receive, send)

    def __eq__(self, other: t.Any) -> bool:
        return super().__eq__(other) and isinstance(other, Mount)

    def build(self, app: t.Optional["Flama"] = None) -> None:
        """Build step for routes.

        Just build the parameters' descriptor part of RouteParametersMixin.

        :param app: Flama app.
        """
        if app and types.is_flama_instance(self.app):
            self.app.router.components = Components(self.app.router.components + app.components)

        if root := (self.app if types.is_flama_instance(self.app) else app):
            for route in self.routes:
                route.build(root)

    def match(self, scope: types.Scope) -> BaseRoute.Match:
        """Check if this route matches with given scope.

        :param scope: ASGI scope.
        :return: Match.
        """
        if scope["type"] not in ("http", "websocket"):
            return self.Match.none

        return (
            self.Match.full
            if self.path.match(scope["path"]).match in (self.path.Match.exact, self.path.Match.partial)
            else self.Match.none
        )

    async def handle(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        """Performs a request by calling the app of this route.

        :param scope: ASGI scope.
        :param receive: ASGI receive event.
        :param send: ASGI send event
        """
        await concurrency.run(self.app, scope, receive, send)

    def route_scope(self, scope: types.Scope) -> types.Scope:
        """Build route scope from given scope.

        It generates an updated scope parameters for the route:
        * app: The app of this mount point. If it's mounting a Flama app, it will replace the app with this one
        * path_params: The matched path parameters of this mount point
        * endpoint: The endpoint of this mount point
        * root_path: The root path of this mount point (if it's mounting a Flama app, it will be empty)
        * path: The remaining path to be matched

        :param scope: ASGI scope.
        :return: Route scope.
        """
        is_flama = types.is_flama_instance(self.app)
        result = {"app": self.app if is_flama else scope["app"]}

        if "path" in scope:
            path = scope["path"]
            match = self.path.match(path)
            result.update(
                {
                    "root_path": "" if is_flama else str(url.Path(scope.get("root_path", "")) / (match.matched or "")),
                    "path": str(url.Path("/") / (match.unmatched or "")),
                }
            )

        return types.Scope(result)

    def resolve_url(self, name: str, **params: t.Any) -> url.URL:
        """Builds URL path for given name and params.

        :param name: Route name.
        :param params: Path params.
        :return: URL path.
        """
        try:
            app_name, rest = name.split(":", 1)
        except ValueError:
            app_name, rest = name, None

        if app_name != self.name:
            raise exceptions.NotFoundException(params=params, name=name)

        try:
            path = self.path.build(**params)
        except ValueError:
            raise exceptions.NotFoundException(params=params, name=name)

        if rest is None:
            return url.URL(path=path.path, scheme="http")

        for route in self.routes:
            try:
                route_url = route.resolve_url(rest, **path.unused)
                return url.URL(path=path.path / route_url.path, scheme="http")
            except exceptions.NotFoundException:
                pass

        raise exceptions.NotFoundException(params=params, name=name)

    @property
    def routes(self) -> list[BaseRoute]:
        """Get all routes registered in this Mount.

        :return: List of routes.
        """
        return getattr(self.app, "routes", [])



import typing as t
from datetime import datetime

import pydantic

import flama
from flama import Flama, schemas
from flama.models import ModelResource
from flama.resources import resource_method

app = Flama(
    openapi={
        "info": {
            "title": "Flama ML",
            "version": "0.1.0",
            "description": "Machine learning API using Flama 🔥",
        }
    },
    docs="/docs/",
)


class X(pydantic.BaseModel):
    input: list[t.Any] = pydantic.Field(title="input", description="Model input")


class Y(pydantic.BaseModel):
    output: list[t.Any] = pydantic.Field(title="output", description="Model output")


app.schema.register_schema("X", X)
app.schema.register_schema("Y", Y)


class MySKModel(ModelResource):
    # special names:
    name = "sk_model"
    verbose_name = "My ScikitLearn Model"
    model_path = "sklearn_model.flm"

    # custom attributes
    info = {
        "model_version": "1.0.0",
        "library_version": "1.0.2",
    }

    @resource_method("/predict/", methods=["POST"], name="model-predict")
    def predict(
        self, data: t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(X)]
    ) -> t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(Y)]:
        """
        tags:
            - My ScikitLearn Model
        summary:
            Run predict method.
        description:
            Runs a prediction using the model from this resource.
        responses:
            200:
                description: ML model prediction.
        """
        return {"output": self.model.predict(data["input"])}

    @resource_method("/inspect/", methods=["GET"], name="model-inspect-model")
    def inspect_model(self):
        """
        tags:
            - My ScikitLearn Model
        summary:
            Run model inspect method.
        description:
            Run the built-in inspect method of the model.
        responses:
            200:
                description: Model inspection.
        """
        return {"params": self.model.inspect()}

    @resource_method("/metadata/", methods=["GET"], name="metadata-method")
    def metadata(self):
        """
        tags:
            - My ScikitLearn Model
        summary:
            Get metadata info.
        description:
            Return metadata info about the model, showing both the bui
        responses:
            200:
                description: ML model metadata.
        """
        return {
            "metadata": {
                "built-in": {
                    "verbose_name": self._meta.verbose_name,
                    "name": self._meta.name,
                },
                "custom": {
                    **self.info,
                    "date": datetime.now().date(),
                    "time": datetime.now().time(),
                },
            }
        }


app.models.add_model_resource(path="/model", resource=MySKModel)

if __name__ == "__main__":
    flama.run(flama_app="__main__:app", server_host="0.0.0.0", server_port=8080, server_reload=True)



import datetime
import typing

import pytest

from flama.injection.exceptions import ComponentNotFound


class TestCaseParamsValidation:
    @pytest.fixture(scope="function", autouse=True)
    def add_path_endpoints(self, app):
        @app.route("/str-path-param/{param}/")
        def str_path_param(param: str):
            assert isinstance(param, str)
            return {"param": param}

        @app.route("/int-path-param/{param}/")
        def int_path_param(param: int):
            assert isinstance(param, int)
            return {"param": param}

        @app.route("/float-path-param/{param}/")
        def float_path_param(param: float):
            assert isinstance(param, float)
            return {"param": param}

        @app.route("/bool-path-param/{param}/")
        def bool_path_param(param: bool):
            assert isinstance(param, bool)
            return {"param": param}

        @app.route("/datetime-path-param/{param}/")
        def datetime_path_param(param: datetime.datetime):
            assert isinstance(param, datetime.datetime)
            return {"param": param}

        @app.route("/date-path-param/{param}/")
        def date_path_param(param: datetime.date):
            assert isinstance(param, datetime.date)
            return {"param": param}

        @app.route("/time-path-param/{param}/")
        def time_path_param(param: datetime.time):
            assert isinstance(param, datetime.time)
            return {"param": param}

    @pytest.fixture(scope="function", autouse=True)
    def add_query_endpoints(self, app):
        @app.route("/str-query-param/")
        def str_query_param(param: str):
            assert isinstance(param, str)
            return {"param": param}

        @app.route("/int-query-param/")
        def int_query_param(param: int):
            assert isinstance(param, int)
            return {"param": param}

        @app.route("/float-query-param/")
        def float_query_param(param: float):
            assert isinstance(param, float)
            return {"param": param}

        @app.route("/bool-query-param/")
        def bool_query_param(param: bool):
            assert isinstance(param, bool)
            return {"param": param}

        @app.route("/datetime-query-param/")
        def datetime_query_param(param: datetime.datetime):
            assert isinstance(param, datetime.datetime)
            return {"param": param}

        @app.route("/date-query-param/")
        def date_query_param(param: datetime.date):
            assert isinstance(param, datetime.date)
            return {"param": param}

        @app.route("/time-query-param/")
        def time_query_param(param: datetime.time):
            assert isinstance(param, datetime.time)
            return {"param": param}

    @pytest.fixture(scope="function", autouse=True)
    def add_query_with_default_endpoints(self, app):
        @app.route("/str-query-param-with_default/")
        def str_query_param_with_default(param: str = "Foo"):
            assert isinstance(param, str)
            return {"param": param}

        @app.route("/int-query-param-with_default/")
        def int_query_param_with_default(param: int = 0):
            assert isinstance(param, int)
            return {"param": param}

        @app.route("/float-query-param-with_default/")
        def float_query_param_with_default(param: float = 0.0):
            assert isinstance(param, float)
            return {"param": param}

        @app.route("/bool-query-param-with_default/")
        def bool_query_param_with_default(param: bool = False):
            assert isinstance(param, bool)
            return {"param": param}

        @app.route("/datetime-query-param-with_default/")
        def datetime_query_param_with_default(param: datetime.datetime = datetime.datetime(2018, 1, 1, 0, 0, 0)):
            assert isinstance(param, datetime.datetime)
            return {"param": param}

        @app.route("/date-query-param-with_default/")
        def date_query_param_with_default(param: datetime.date = datetime.date(2018, 1, 1)):
            assert isinstance(param, datetime.date)
            return {"param": param}

        @app.route("/time-query-param-with_default/")
        def time_query_param_with_default(param: datetime.time = datetime.time(0, 0, 0)):
            assert isinstance(param, datetime.time)
            return {"param": param}

    @pytest.fixture(scope="function", autouse=True)
    def add_query_optional_endpoints(self, app):
        @app.route("/str-query-param-optional/")
        def str_query_param_optional(param: typing.Optional[str] = None):
            assert param is None
            return {"param": param}

        @app.route("/int-query-param-optional/")
        def int_query_param_optional(param: typing.Optional[int] = None):
            assert param is None
            return {"param": param}

        @app.route("/float-query-param-optional/")
        def float_query_param_optional(param: typing.Optional[float] = None):
            assert param is None
            return {"param": param}

        @app.route("/bool-query-param-optional/")
        def bool_query_param_optional(param: typing.Optional[bool] = None):
            assert param is None
            return {"param": param}

        @app.route("/datetime-query-param-optional/")
        def datetime_query_param_optional(param: typing.Optional[datetime.datetime] = None):
            assert param is None
            return {"param": param}

        @app.route("/date-query-param-optional/")
        def date_query_param_optional(param: typing.Optional[datetime.date] = None):
            assert param is None
            return {"param": param}

        @app.route("/time-query-param-optional/")
        def time_query_param_optional(param: typing.Optional[datetime.time] = None):
            assert param is None
            return {"param": param}

    @pytest.fixture(scope="function", autouse=True)
    def add_other_endpoints(self, app):
        @app.route("/empty/", methods=["POST"])
        def empty(foo):
            return {}

    @pytest.mark.parametrize(
        ["url", "value"],
        [
            pytest.param("/str-path-param/123/", "123", id="str-path-param"),
            pytest.param("/int-path-param/123/", 123, id="int-path-param"),
            pytest.param("/float-path-param/123.321/", 123.321, id="float-path-param"),
            pytest.param("/bool-path-param/true/", True, id="float-path-param"),
            pytest.param(
                "/datetime-path-param/2018-01-01T00:00:00+00:00/", "2018-01-01T00:00:00+00:00", id="datetime-path-param"
            ),
            pytest.param("/date-path-param/2018-01-01/", "2018-01-01", id="date-path-param"),
            pytest.param("/time-path-param/00:00:00/", "00:00:00", id="time-path-param"),
        ],
    )
    async def test_path_param(self, url, value, client):
        response = await client.get(url)
        assert response.json() == {"param": value}

    @pytest.mark.parametrize(
        ["url", "value"],
        [
            pytest.param("/str-query-param/", "123", id="str-query-param"),
            pytest.param("/int-query-param/", 123, id="int-query-param"),
            pytest.param("/float-query-param/", 123.321, id="float-query-param"),
            pytest.param("/bool-query-param/", True, id="bool-query-param"),
            pytest.param("/datetime-query-param/", "2018-01-01T00:00:00", id="datetime-query-param"),
            pytest.param("/date-query-param/", "2018-01-01", id="date-query-param"),
            pytest.param("/time-query-param/", "00:00:00", id="time-query-param"),
        ],
    )
    async def test_query_param(self, url, value, client):
        response = await client.get(url, params={"param": value})
        assert response.json() == {"param": value}

    @pytest.mark.parametrize(
        ["url", "value"],
        [
            pytest.param("/str-query-param-with_default/", "Foo", id="str-query-param-with-default"),
            pytest.param("/int-query-param-with_default/", 0, id="int-query-param-with-default"),
            pytest.param("/float-query-param-with_default/", 0.0, id="float-query-param-with-default"),
            pytest.param("/bool-query-param-with_default/", False, id="bool-query-param-with-default"),
            pytest.param(
                "/datetime-query-param-with_default/", "2018-01-01T00:00:00", id="datetime-query-param-with-default"
            ),
            pytest.param("/date-query-param-with_default/", "2018-01-01", id="date-query-param-with-default"),
            pytest.param("/time-query-param-with_default/", "00:00:00", id="time-query-param-with-default"),
        ],
    )
    async def test_query_param_with_default(self, url, value, client):
        response = await client.get(url, params={"param": value})
        assert response.json() == {"param": value}

    @pytest.mark.parametrize(
        "url",
        [
            pytest.param("/str-query-param-optional/", id="str-query-param-optional"),
            pytest.param("/int-query-param-optional/", id="int-query-param-optional"),
            pytest.param("/float-query-param-optional/", id="float-query-param-optional"),
            pytest.param("/bool-query-param-optional/", id="bool-query-param-optional"),
            pytest.param("/datetime-query-param-optional/", id="datetime-query-param-optional"),
            pytest.param("/date-query-param-optional/", id="date-query-param-optional"),
            pytest.param("/time-query-param-optional/", id="time-query-param-optional"),
        ],
    )
    async def test_query_param_optional(self, url, client):
        response = await client.get(url)
        assert response.json() == {"param": None}

    async def test_wrong_query_param(self, client):
        response = await client.get("/int-query-param/?param=foo")
        assert response.status_code == 400

    async def test_wrong_path_param(self, client):
        response = await client.get("/int-path-param/foo/")
        assert response.status_code == 400

    async def test_no_type_param(self, client):
        with pytest.raises(ComponentNotFound, match="No component able to handle parameter 'foo' for function 'empty'"):
            await client.post("/empty/")



import pytest

from flama import Flama
from flama.client import Client
from flama.ddd.repositories import HTTPRepository
from flama.ddd.repositories.sqlalchemy import SQLAlchemyRepository
from flama.exceptions import ApplicationError
from flama.resources.workers import FlamaWorker
from flama.sqlalchemy import SQLAlchemyModule


class TestCaseFlamaWorker:
    @pytest.fixture(scope="function")
    async def app(self):
        app_ = Flama(schema=None, docs=None, modules={SQLAlchemyModule("sqlite+aiosqlite://")})
        async with Client(app_):
            yield app_

    @pytest.fixture(scope="function")
    def worker(self):
        class FooWorker(FlamaWorker): ...

        return FooWorker()

    @pytest.fixture(scope="function")
    def sqlalchemy_repository(self):
        class FooRepository(SQLAlchemyRepository): ...

        return FooRepository

    @pytest.fixture(scope="function")
    def http_repository(self):
        class BarRepository(HTTPRepository): ...

        return BarRepository

    def test_init(self):
        worker = FlamaWorker()

        assert not worker._resources_repositories.registered

    async def test_async_context(self, app, worker, http_repository, sqlalchemy_repository):
        worker.app = app
        worker.add_repository("foo", sqlalchemy_repository)
        worker.add_repository("bar", http_repository)

        with pytest.raises(ApplicationError, match="Repositories not initialized"):
            worker.repositories

        async with worker:
            assert worker.repositories == {
                "foo": sqlalchemy_repository(worker.connection),
                "bar": http_repository(worker.connection),
            }

        worker.remove_repository("bar")

        async with worker:
            assert worker.repositories == {"foo": sqlalchemy_repository(worker.connection)}



from typesystem import Definitions, Reference, Schema, fields

__all__ = [
    "APIError",
    "DropCollection",
    "LimitOffsetMeta",
    "LimitOffset",
    "PageNumberMeta",
    "PageNumber",
    "MLModelInput",
    "MLModelOutput",
    "SCHEMAS",
]

SCHEMAS = Definitions()

APIError = Schema(
    title="APIError",
    fields={
        "status_code": fields.Integer(title="status_code", description="HTTP status code"),
        "detail": fields.Union(any_of=[fields.Object(), fields.String()], title="detail", description="Error detail"),
        "error": fields.String(title="type", description="Exception or error type", allow_null=True),
    },
)
SCHEMAS["flama.APIError"] = APIError

DropCollection = Schema(
    title="DropCollection",
    fields={
        "deleted": fields.Integer(title="deleted", description="Number of deleted elements"),
    },
)
SCHEMAS["flama.DropCollection"] = DropCollection

LimitOffsetMeta = Schema(
    title="LimitOffsetMeta",
    fields={
        "limit": fields.Integer(title="limit", description="Number of retrieved items"),
        "offset": fields.Integer(title="offset", description="Collection offset"),
        "count": fields.Integer(title="count", description="Total number of items", allow_null=True),
    },
)
SCHEMAS["flama.LimitOffsetMeta"] = LimitOffsetMeta

LimitOffset = Schema(
    title="LimitOffset",
    fields={
        "meta": Reference(
            to="flama.LimitOffsetMeta", definitions=SCHEMAS, title="meta", description="Pagination metadata"
        ),
        "data": fields.Array(title="data", description="Paginated data"),
    },
)
SCHEMAS["flama.LimitOffset"] = LimitOffset

PageNumberMeta = Schema(
    title="PageNumberMeta",
    fields={
        "page": fields.Integer(title="page", description="Current page number"),
        "page_size": fields.Integer(title="page_size", description="Page size"),
        "count": fields.Integer(title="count", description="Total number of items", allow_null=True),
    },
)
SCHEMAS["flama.PageNumberMeta"] = PageNumberMeta

PageNumber = Schema(
    title="PageNumber",
    fields={
        "meta": Reference(
            to="flama.PageNumberMeta", definitions=SCHEMAS, title="meta", description="Pagination metadata"
        ),
        "data": fields.Array(title="data", description="Paginated data"),
    },
)
SCHEMAS["flama.PageNumber"] = PageNumber

MLModelInput = Schema(
    title="MLModelInput",
    fields={
        "input": fields.Array(title="input", description="Model input"),
    },
)
SCHEMAS["flama.MLModelInput"] = MLModelInput

MLModelOutput = Schema(
    title="MLModelOutput",
    fields={
        "output": fields.Array(title="output", description="Model output"),
    },
)
SCHEMAS["flama.MLModelOutput"] = MLModelOutput



import datetime
import typing as t
import uuid

import marshmallow
import pydantic
import pytest
import sqlalchemy
import typesystem
import typesystem.fields
from sqlalchemy.dialects import postgresql

from flama.applications import Flama
from flama.resources.crud import CRUDResource
from flama.resources.routing import ResourceRoute, resource_method
from flama.resources.workers import FlamaWorker
from flama.schemas import SchemaMetadata, SchemaType
from flama.sqlalchemy import SQLAlchemyModule
from tests.conftest import DATABASE_URL


@pytest.fixture(scope="function")
def app(app):
    # Remove schema and docs endpoint from base fixture
    return Flama(
        schema=None, docs=None, modules={SQLAlchemyModule("sqlite+aiosqlite://")}, schema_library=app.schema_library
    )


@pytest.fixture(scope="function", autouse=True)
def add_resources(app, resource):
    app.resources.add_resource("/puppy/", resource)

    yield

    app.resources.remove_repository(resource._meta.name)


@pytest.fixture(scope="function")
def puppy():
    return {"name": "canna", "age": 2, "owner": "Perdy"}


@pytest.fixture(scope="function")
def another_puppy():
    return {"name": "sandy", "age": 6, "owner": "Perdy"}


class TestCaseCRUDResource:
    @pytest.fixture(scope="function")
    def resource(self, puppy_model, puppy_schema):
        class PuppyResource(CRUDResource):
            name = "puppy"
            verbose_name = "Puppy"

            model = puppy_model
            input_schema = puppy_schema
            output_schema = puppy_schema

            @resource_method("/", methods=["GET"], name="list", pagination="page_number")
            async def list(
                self,
                worker: FlamaWorker,
                order_by: t.Optional[str] = None,
                order_direction: str = "asc",
                name: t.Optional[str] = None,
                custom_id__le: t.Optional[int] = None,
                **kwargs,
            ) -> t.Annotated[list[SchemaType], SchemaMetadata(puppy_schema)]:
                """
                description: Custom list method with filtering by name.
                """
                clauses = []

                if custom_id__le is not None:
                    clauses.append(self.model.c.custom_id <= custom_id__le)

                filters = {}

                if name is not None:
                    filters["name"] = name

                async with worker:
                    return [
                        x
                        async for x in worker.repositories[self._meta.name].list(
                            *clauses, order_by=order_by, order_direction=order_direction, **filters
                        )
                    ]

        return PuppyResource()

    @pytest.fixture(scope="function")
    async def custom_id_datetime_model(self, app):
        table = sqlalchemy.Table(
            "custom_id_datetime",
            app.sqlalchemy.metadata,
            sqlalchemy.Column("custom_id", sqlalchemy.DateTime, primary_key=True),
            sqlalchemy.Column("name", sqlalchemy.String),
        )

        async with app.sqlalchemy.engine.begin() as connection:
            await connection.run_sync(app.sqlalchemy.metadata.create_all, tables=[table])

        yield table

        async with app.sqlalchemy.engine.begin() as connection:
            await connection.run_sync(app.sqlalchemy.metadata.drop_all, tables=[table])

    @pytest.fixture(scope="function")
    def custom_id_datetime_schema(self, app):
        from flama import schemas

        if schemas.lib == pydantic:
            schema_ = pydantic.create_model("CustomIDDatetime", custom_id=(datetime.datetime, ...), name=(str, ...))
        elif schemas.lib == typesystem:
            schema_ = typesystem.Schema(
                title="CustomIDDatetime",
                fields={
                    "custom_id": typesystem.fields.DateTime(),
                    "name": typesystem.fields.String(),
                },
            )
        elif schemas.lib == marshmallow:
            schema_ = type(
                "CustomIDDatetime",
                (marshmallow.Schema,),
                {
                    "custom_id": marshmallow.fields.DateTime(),
                    "name": marshmallow.fields.String(),
                },
            )
        else:
            raise ValueError("Wrong schema lib")

        return schema_

    @pytest.fixture(scope="function")
    def custom_id_datetime_resource(self, custom_id_datetime_model, custom_id_datetime_schema, app):
        class CustomUUIDResource(CRUDResource):
            model = custom_id_datetime_model
            schema = custom_id_datetime_schema

            name = "custom_id_datetime"

        app.resources.add_resource("/custom_id_datetime/", CustomUUIDResource)

        yield CustomUUIDResource

        app.resources.remove_repository(CustomUUIDResource._meta.name)

    @pytest.fixture(scope="function")
    async def custom_id_uuid_model(self, app):
        table = sqlalchemy.Table(
            "custom_id_uuid",
            app.database.metadata,
            sqlalchemy.Column("custom_id", postgresql.UUID, primary_key=True),
            sqlalchemy.Column("name", sqlalchemy.String),
        )

        async with app.database.engine.begin() as connection:
            await connection.run_sync(app.database.metadata.create_all, tables=[table])

        yield table

        async with app.database.engine.begin() as connection:
            await connection.run_sync(app.database.metadata.drop_all, tables=[table])

    @pytest.fixture(scope="function")
    def custom_id_uuid_schema(self, app):
        from flama import schemas

        if schemas.lib == pydantic:
            schema_ = pydantic.create_model("CustomIDUUID", custom_id=(uuid.UUID, ...), name=(str, ...))
        elif schemas.lib == typesystem:
            schema_ = typesystem.Schema(
                title="CustomIDUUID",
                fields={
                    "custom_id": typesystem.fields.UUID(),
                    "name": typesystem.fields.String(),
                },
            )
        elif schemas.lib == marshmallow:
            schema_ = type(
                "CustomIDUUID",
                (marshmallow.Schema,),
                {
                    "custom_id": marshmallow.fields.UUID(),
                    "name": marshmallow.fields.String(),
                },
            )
        else:
            raise ValueError("Wrong schema lib")

        return schema_

    @pytest.fixture(scope="function")
    def custom_id_uuid_resource(self, custom_id_uuid_model, custom_id_uuid_schema, app):
        class CustomUUIDResource(CRUDResource):
            model = custom_id_uuid_model
            schema = custom_id_uuid_schema

            name = "custom_id_uuid"

        app.resources.add_resource("/custom_id_datetime/", CustomUUIDResource)

        yield CustomUUIDResource

        del app.resources.worker._repositories[CustomUUIDResource._meta.name]

    def test_crud_resource(self, resource, app):
        expected_routes = [
            ("/", resource.list, {"GET", "HEAD"}, "list"),
            ("/", resource.create, {"POST"}, "create"),
            ("/{resource_id}/", resource.retrieve, {"GET", "HEAD"}, "retrieve"),
            ("/{resource_id}/", resource.update, {"PUT"}, "update"),
            ("/{resource_id}/", resource.partial_update, {"PATCH"}, "partial-update"),
            ("/{resource_id}/", resource.delete, {"DELETE"}, "delete"),
            ("/", resource.replace, {"PUT"}, "replace"),
            ("/", resource.partial_replace, {"PATCH"}, "partial-replace"),
            ("/", resource.drop, {"DELETE"}, "drop"),
        ]

        assert hasattr(resource, "create")
        assert hasattr(resource, "retrieve")
        assert hasattr(resource, "update")
        assert hasattr(resource, "partial_update")
        assert hasattr(resource, "delete")
        assert hasattr(resource, "list")
        assert hasattr(resource, "replace")
        assert hasattr(resource, "partial_replace")
        assert hasattr(resource, "drop")
        assert len(app.routes) == 1
        assert isinstance(app.routes[0], ResourceRoute)
        resource_route = app.routes[0]
        assert [(i.path, i.endpoint, i.methods, i.name) for i in resource_route.routes] == expected_routes

    async def test_create(self, client, puppy):
        expected_puppy_id = 1
        expected_puppy = puppy.copy()
        expected_puppy["custom_id"] = expected_puppy_id

        # Successfully create a new record
        response = await client.request("post", "/puppy/", json=puppy)
        assert response.status_code == 201, response.json()
        created_puppy = response.json()
        assert created_puppy == expected_puppy

        # List all the existing records
        response = await client.request("get", f"/puppy/{expected_puppy_id}/")
        assert response.status_code == 200, response.json()
        assert response.json() == expected_puppy

    async def test_create_already_exists(self, client, puppy):
        expected_puppy_id = 1
        expected_puppy = puppy.copy()
        expected_puppy["custom_id"] = expected_puppy_id

        # Create a new record
        response = await client.request("post", "/puppy/", json=puppy)
        assert response.status_code == 201, response.json()
        created_puppy = response.json()
        assert created_puppy == expected_puppy

        # Fails if resource exists
        response = await client.request("post", "/puppy/", json=expected_puppy)
        assert response.status_code == 400, response.json()

    async def test_create_wrong_data(self, client, puppy):
        wrong_puppy = puppy.copy()
        wrong_puppy["age"] = "wrong"

        # Fails if wrong input data
        response = await client.request("post", "/puppy/", json=wrong_puppy)
        assert response.status_code == 400, response.json()

    async def test_retrieve(self, client, puppy):
        expected_puppy_id = 1
        expected_result = puppy.copy()
        expected_result["custom_id"] = expected_puppy_id

        # Successfully create a new record
        response = await client.request("post", "/puppy/", json=puppy)
        assert response.status_code == 201, response.json()
        created_puppy = response.json()
        assert created_puppy == expected_result

        # Retrieve same record
        response = await client.request("get", f"/puppy/{expected_puppy_id}/")
        assert response.status_code == 200, response.json()
        assert response.json() == expected_result

    async def test_retrieve_not_found(self, client):
        # Retrieve wrong record
        response = await client.request("get", "/puppy/42/")
        assert response.status_code == 404, response.json()

    async def test_retrieve_wrong_id_type(self, client):
        # Retrieve wrong record
        response = await client.request("get", "/puppy/foo/")
        assert response.status_code == 400, response.json()

    async def test_update(self, client, puppy, another_puppy):
        expected_puppy_id = 1
        created_puppy = puppy.copy()
        created_puppy["custom_id"] = expected_puppy_id
        expected_puppy = another_puppy.copy()
        another_puppy.pop("owner")
        expected_puppy["custom_id"] = expected_puppy_id
        expected_puppy["owner"] = None  # Replaced by default

        # Successfully create a new record
        response = await client.request("post", "/puppy/", json=puppy)
        assert response.status_code == 201, response.json()
        assert response.json() == created_puppy

        # Update record
        response = await client.request("put", f"/puppy/{expected_puppy_id}/", json=another_puppy)
        assert response.status_code == 200, response.json()
        assert response.json() == expected_puppy

        # List all the existing records
        response = await client.request("get", f"/puppy/{expected_puppy_id}/")
        assert response.status_code == 200, response.json()
        assert response.json() == expected_puppy

    async def test_update_not_found(self, client, puppy):
        # Update wrong record
        response = await client.request("put", "/puppy/42/", json=puppy)
        assert response.status_code == 404, response.json()

    async def test_update_wrong_id_type(self, client, puppy):
        # Update wrong record
        response = await client.request("put", "/puppy/foo/", json=puppy)
        assert response.status_code == 400, response.json()

    async def test_update_wrong_data(self, client, puppy):
        expected_puppy_id = 1
        created_puppy = puppy.copy()
        created_puppy["custom_id"] = expected_puppy_id
        wrong_puppy = puppy.copy()
        wrong_puppy["age"] = "wrong"

        # Successfully create a new record
        response = await client.request("post", "/puppy/", json=puppy)
        assert response.status_code == 201, response.json()
        assert response.json() == created_puppy

        # Update record
        response = await client.request("put", f"/puppy/{expected_puppy_id}/", json=wrong_puppy)
        assert response.status_code == 400, response.json()

    async def test_partial_update(self, client, puppy, another_puppy):
        expected_puppy_id = 1
        created_puppy = puppy.copy()
        created_puppy["custom_id"] = expected_puppy_id
        expected_puppy = another_puppy.copy()
        another_puppy.pop("owner")
        expected_puppy["custom_id"] = expected_puppy_id
        expected_puppy["owner"] = created_puppy["owner"]  # Not replaced

        # Successfully create a new record
        response = await client.request("post", "/puppy/", json=puppy)
        assert response.status_code == 201, response.json()
        assert response.json() == created_puppy

        # Update record
        response = await client.request("patch", f"/puppy/{expected_puppy_id}/", json=another_puppy)
        assert response.status_code == 200, response.json()
        assert response.json() == expected_puppy

        # List all the existing records
        response = await client.request("get", f"/puppy/{expected_puppy_id}/")
        assert response.status_code == 200, response.json()
        assert response.json() == expected_puppy

    async def test_partial_update_not_found(self, client, puppy):
        # Update wrong record
        response = await client.request("patch", "/puppy/42/", json=puppy)
        assert response.status_code == 404, response.json()

    async def test_partial_update_wrong_id_type(self, client, puppy):
        # Update wrong record
        response = await client.request("patch", "/puppy/foo/", json=puppy)
        assert response.status_code == 400, response.json()

    async def test_partial_update_wrong_data(self, client, puppy):
        expected_puppy_id = 1
        created_puppy = puppy.copy()
        created_puppy["custom_id"] = expected_puppy_id
        wrong_puppy = puppy.copy()
        wrong_puppy["age"] = "wrong"

        # Successfully create a new record
        response = await client.request("post", "/puppy/", json=puppy)
        assert response.status_code == 201, response.json()
        assert response.json() == created_puppy

        # Update record
        response = await client.request("patch", f"/puppy/{expected_puppy_id}/", json=wrong_puppy)
        assert response.status_code == 400, response.json()

    async def test_delete(self, client, puppy):
        expected_puppy_id = 1
        expected_puppy = puppy.copy()
        expected_puppy["custom_id"] = expected_puppy_id

        # Successfully create a new record
        response = await client.request("post", "/puppy/", json=puppy)
        assert response.status_code == 201, response.json()
        created_puppy = response.json()
        assert created_puppy == expected_puppy

        # Retrieve same record
        response = await client.request("get", f"/puppy/{expected_puppy_id}/")
        assert response.status_code == 200, response.json()
        assert response.json() == expected_puppy

        # Delete record
        response = await client.request("delete", f"/puppy/{expected_puppy_id}/")
        assert response.status_code == 204, response.json()

        # Retrieve deleted record
        response = await client.request("get", f"/puppy/{expected_puppy_id}/")
        assert response.status_code == 404, response.json()

    async def test_delete_wrong_id_type(self, client):
        # Delete wrong record
        response = await client.request("delete", "/puppy/foo/")
        assert response.status_code == 400, response.json()

    async def test_delete_not_found(self, client, puppy):
        # Delete wrong record
        response = await client.request("delete", "/puppy/42/", json=puppy)
        assert response.status_code == 404, response.json()

    @pytest.mark.skipif(not DATABASE_URL.startswith("postgresql"), reason="Only valid for PostgreSQL backend")
    async def test_id_uuid(self, app, client, custom_id_uuid_resource):
        data = {"custom_id": str(uuid.uuid4()), "name": "foo"}
        expected_result = data.copy()

        # Successfully create a new record
        response = await client.request("post", "/custom_id_uuid/", json=data)
        assert response.status_code == 201, response.content
        assert response.json() == expected_result, response.json()

        # Retrieve same record
        response = await client.request("get", f"/custom_id_uuid/{data['custom_id']}/")
        assert response.status_code == 200, response.json()
        assert response.json() == expected_result

    async def test_id_datetime(self, client, app, custom_id_datetime_resource):
        data = {"custom_id": "2018-01-01T00:00:00", "name": "foo"}
        expected_result = data.copy()

        # Successfully create a new record
        response = await client.request("post", "/custom_id_datetime/", json=data)
        assert response.status_code == 201, response.content
        assert response.json() == expected_result, response.json()

        # Retrieve same record
        response = await client.request("get", f"/custom_id_datetime/{data['custom_id']}/")
        assert response.status_code == 200, response.json()
        assert response.json() == expected_result

    async def test_list(self, client, puppy, another_puppy):
        # Successfully create a new record
        response = await client.request("post", "/puppy/", json=puppy)
        assert response.status_code == 201, response.json()

        # Successfully create another new record
        response = await client.request("post", "/puppy/", json=another_puppy)
        assert response.status_code == 201, response.json()

        # List all the existing records
        response = await client.request("get", "/puppy/")
        assert response.status_code == 200, response.json()
        assert response.json()["data"] == [{"custom_id": 1, **puppy}, {"custom_id": 2, **another_puppy}]

    async def test_list_order(self, client, puppy, another_puppy):
        # Successfully create a new record
        response = await client.request("post", "/puppy/", json=puppy)
        assert response.status_code == 201, response.json()

        # Successfully create another new record
        response = await client.request("post", "/puppy/", json=another_puppy)
        assert response.status_code == 201, response.json()

        # List all the existing records
        response = await client.request("get", "/puppy/", params={"order_by": "name", "order_direction": "desc"})
        assert response.status_code == 200, response.json()
        assert response.json()["data"] == [{"custom_id": 2, **another_puppy}, {"custom_id": 1, **puppy}]

    async def test_list_filter(self, client, puppy, another_puppy):
        # Successfully create a new record
        response = await client.request("post", "/puppy/", json=puppy)
        assert response.status_code == 201, response.json()

        # Successfully create another new record
        response = await client.request("post", "/puppy/", json=another_puppy)
        assert response.status_code == 201, response.json()

        # Filter and found something
        response = await client.request("get", "/puppy/", params={"name": "canna", "custom_id__le": 1})
        assert response.status_code == 200, response.json()
        assert response.json()["data"] == [{"custom_id": 1, **puppy}]

        # Filter without results
        response = await client.request("get", "/puppy/", params={"name": "canna", "custom_id__le": 0})
        assert response.status_code == 200, response.json()
        assert response.json()["data"] == []

    async def test_replace(self, client, puppy, another_puppy):
        # Successfully create a new record
        response = await client.request("post", "/puppy/", json=puppy)
        assert response.status_code == 201, response.json()

        # Successfully create a new record
        response = await client.request("post", "/puppy/", json=another_puppy)
        assert response.status_code == 201, response.json()

        # List all the existing records
        response = await client.request("get", "/puppy/")
        assert response.status_code == 200, response.json()
        assert response.json()["data"] == [{"custom_id": 1, **puppy}, {"custom_id": 2, **another_puppy}]

        # Replace collection
        response = await client.request(
            "put", "/puppy/", json=[{"custom_id": 2, **puppy}, {"custom_id": 3, **another_puppy}]
        )
        assert response.status_code == 200, response.json()
        assert response.json() == [{"custom_id": 2, **puppy}, {"custom_id": 3, **another_puppy}]

        # List all the existing records
        response = await client.request("get", "/puppy/")
        assert response.status_code == 200, response.json()
        assert response.json()["data"] == [{"custom_id": 2, **puppy}, {"custom_id": 3, **another_puppy}]

    async def test_replace_wrong_data(self, client, puppy):
        wrong_puppy = puppy.copy()
        wrong_puppy["age"] = "wrong"

        # Successfully create a new record
        response = await client.request("post", "/puppy/", json=puppy)
        assert response.status_code == 201, response.json()

        # List all the existing records
        response = await client.request("get", "/puppy/")
        assert response.status_code == 200, response.json()
        assert response.json()["data"] == [{"custom_id": 1, **puppy}]

        # Fails if wrong input data
        response = await client.request("put", "/puppy/", json=[{"custom_id": 2, **wrong_puppy}])
        assert response.status_code == 400, response.json()

        # Collection remains the same
        response = await client.request("get", "/puppy/")
        assert response.status_code == 200, response.json()
        assert response.json()["data"] == [{"custom_id": 1, **puppy}]

    async def test_partial_replace(self, client, puppy, another_puppy):
        # Successfully create a new record
        response = await client.request("post", "/puppy/", json=puppy)
        assert response.status_code == 201, response.json()

        # Successfully create a new record
        response = await client.request("post", "/puppy/", json=another_puppy)
        assert response.status_code == 201, response.json()

        # List all the existing records
        response = await client.request("get", "/puppy/")
        assert response.status_code == 200, response.json()
        assert response.json()["data"] == [{"custom_id": 1, **puppy}, {"custom_id": 2, **another_puppy}]

        # Partial replace collection
        response = await client.request(
            "patch", "/puppy/", json=[{"custom_id": 2, **puppy}, {"custom_id": 3, **another_puppy}]
        )
        assert response.status_code == 200, response.json()
        assert response.json() == [{"custom_id": 2, **puppy}, {"custom_id": 3, **another_puppy}]

        # List all the existing records
        response = await client.request("get", "/puppy/")
        assert response.status_code == 200, response.json()
        assert response.json()["data"] == [
            {"custom_id": 1, **puppy},
            {"custom_id": 2, **puppy},
            {"custom_id": 3, **another_puppy},
        ]

    async def test_partial_replace_wrong_data(self, client, puppy):
        wrong_puppy = puppy.copy()
        wrong_puppy["age"] = "wrong"

        # Successfully create a new record
        response = await client.request("post", "/puppy/", json=puppy)
        assert response.status_code == 201, response.json()

        # List all the existing records
        response = await client.request("get", "/puppy/")
        assert response.status_code == 200, response.json()
        assert response.json()["data"] == [{"custom_id": 1, **puppy}]

        # Partial replace collection
        response = await client.request("patch", "/puppy/", json=[{"custom_id": 2, **wrong_puppy}])
        assert response.status_code == 400, response.json()

        # Collection remains the same
        response = await client.request("get", "/puppy/")
        assert response.status_code == 200, response.json()
        assert response.json()["data"] == [{"custom_id": 1, **puppy}]

    async def test_drop(self, client, puppy, another_puppy):
        # Successfully create a new record
        response = await client.request("post", "/puppy/", json=puppy)
        assert response.status_code == 201, response.json()

        # Successfully create a new record
        response = await client.request("post", "/puppy/", json=another_puppy)
        assert response.status_code == 201, response.json()

        # List all the existing records
        response = await client.request("get", "/puppy/")
        assert response.status_code == 200, response.json()
        assert response.json()["data"] == [{"custom_id": 1, **puppy}, {"custom_id": 2, **another_puppy}]

        # Drop collection
        response = await client.request("delete", "/puppy/", json=[puppy])
        assert response.status_code == 204, response.json()
        assert response.json() == {"deleted": 2}

        # List all the existing records
        response = await client.request("get", "/puppy/")
        assert response.status_code == 200, response.json()
        assert response.json()["data"] == []



import multiprocessing
import threading
import warnings

import pytest

from flama import background, http


def sync_task(event):
    event.set()


async def async_task(event):
    event.set()


@pytest.fixture(scope="function")
def process_event():
    return multiprocessing.Event()


@pytest.fixture(scope="function")
def thread_event():
    return threading.Event()


class TestCaseBackgroundTask:
    @pytest.fixture(params=["sync", "async"])
    def task(self, request):
        return sync_task if request.param == "sync" else async_task

    async def test_background_process_task(self, app, client, task, process_event):
        @app.route("/")
        async def test():
            return http.APIResponse({"foo": "bar"}, background=background.BackgroundProcessTask(task, process_event))

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            response = await client.get("/")

        assert response.status_code == 200
        assert response.json() == {"foo": "bar"}

        assert process_event.wait(5.0)

    async def test_background_thread_task(self, app, client, task, thread_event):
        @app.route("/")
        async def test():
            return http.APIResponse({"foo": "bar"}, background=background.BackgroundThreadTask(task, thread_event))

        response = await client.get("/")
        assert response.status_code == 200
        assert response.json() == {"foo": "bar"}

        assert thread_event.wait(5.0)


class TestCaseBackgroundTasks:
    async def test_background_tasks(self, app, client, process_event, thread_event):
        @app.route("/")
        async def test():
            tasks = background.BackgroundTasks()
            tasks.add_task(background.Concurrency.process, sync_task, process_event)
            tasks.add_task(background.Concurrency.thread, async_task, thread_event)
            return http.APIResponse({"foo": "bar"}, background=tasks)

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            response = await client.get("/")

        assert response.status_code == 200
        assert response.json() == {"foo": "bar"}

        assert process_event.wait(5.0)
        assert thread_event.wait(5.0)



import inspect
import typing as t

from flama import exceptions
from flama.modules import Module
from flama.resources.resource import Resource
from flama.resources.routing import ResourceRoute

if t.TYPE_CHECKING:
    try:
        from flama.ddd.repositories.sqlalchemy import SQLAlchemyTableRepository
        from flama.resources.workers import FlamaWorker
    except exceptions.DependencyNotInstalled:
        ...


__all__ = ["ResourcesModule"]


class ResourcesModule(Module):
    name = "resources"

    def __init__(self, worker: t.Optional["FlamaWorker"] = None):
        super().__init__()
        self.worker = worker

    def add_resource(
        self,
        path: str,
        resource: t.Union[Resource, type[Resource]],
        tags: t.Optional[dict[str, dict[str, t.Any]]] = None,
        *args,
        **kwargs,
    ) -> "Resource":
        """Adds a resource to this application, setting its endpoints.

        :param path: Resource base path.
        :param tags: Tags to add to the resource.
        :param resource: Resource class.
        """
        if inspect.isclass(resource) and issubclass(resource, Resource):
            resource_instance = resource(*args, **kwargs)
        elif isinstance(resource, Resource):
            resource_instance = resource
        else:
            raise ValueError("Wrong resource")

        self.app.mount(mount=ResourceRoute(path, resource_instance, tags))

        return resource_instance

    def resource(self, path: str, tags: t.Optional[dict[str, dict[str, t.Any]]] = None, *args, **kwargs) -> t.Callable:
        """Decorator for Resources classes for adding them to the application.

        :param path: Resource base path.
        :param tags: Tags to add to the resource.
        :return: Decorated resource class.
        """

        def decorator(resource: type[Resource]) -> type[Resource]:
            self.add_resource(path, resource, tags, *args, **kwargs)
            return resource

        return decorator

    def add_repository(self, name: str, repository: type["SQLAlchemyTableRepository"]) -> None:
        """Register a repository.

        :param name: The name of the repository.
        :param repository: The repository class.
        """
        if self.worker:
            self.worker.add_repository(name, repository)

    def remove_repository(self, name: str) -> None:
        """Deregister a repository.

        :param name: The name of the repository.
        """
        if self.worker:
            self.worker.remove_repository(name)



import typing as t

from flama.codecs.base import HTTPCodec

if t.TYPE_CHECKING:
    from flama import http

__all__ = ["MultiPartCodec"]


class MultiPartCodec(HTTPCodec):
    media_type = "multipart/form-data"

    async def decode(self, item: "http.Request", **options) -> t.Any:
        return await item.form()



import logging
import typing as t

from flama import exceptions
from flama.ddd.workers.base import BaseWorker

try:
    from sqlalchemy.ext.asyncio import AsyncConnection, AsyncTransaction
except Exception:  # pragma: no cover
    raise exceptions.DependencyNotInstalled(
        dependency=exceptions.DependencyNotInstalled.Dependency.sqlalchemy, dependant=__name__
    )


__all__ = ["SQLAlchemyWorker"]

logger = logging.getLogger(__name__)


class SQLAlchemyWorker(BaseWorker):
    """Worker for SQLAlchemy.

    It will provide a connection and a transaction to the database and create the repositories for the entities.
    """

    _connection: AsyncConnection
    _transaction: AsyncTransaction

    @property
    def connection(self) -> AsyncConnection:
        """Connection to the database.

        :return: Connection to the database.
        :raises AttributeError: If the connection is not initialized.
        """
        try:
            return self._connection
        except AttributeError:
            raise AttributeError("Connection not initialized")

    @connection.setter
    def connection(self, connection: AsyncConnection) -> None:
        """Set the connection to the database.

        :param connection: Connection to the database.
        """
        self._connection = connection

    @connection.deleter
    def connection(self) -> None:
        """Delete the connection to the database."""
        del self._connection

    @property
    def transaction(self) -> AsyncTransaction:
        """Database transaction.

        :return: Database transaction.
        :raises AttributeError: If the transaction is not started.
        """
        try:
            return self._transaction
        except AttributeError:
            raise AttributeError("Transaction not started")

    @transaction.setter
    def transaction(self, transaction: AsyncTransaction) -> None:
        """Set the transaction.

        :param transaction: Database transaction.
        """
        self._transaction = transaction

    @transaction.deleter
    def transaction(self) -> None:
        """Delete the transaction."""
        del self._transaction

    async def set_up(self) -> None:
        """Open a connection and begin a transaction."""

        self.connection = await self.app.sqlalchemy.open_connection()
        self.transaction = await self.app.sqlalchemy.begin_transaction(self._connection)

    async def tear_down(self, *, rollback: bool = False) -> None:
        """End a transaction and close the connection.

        :param rollback: If the transaction should be rolled back.
        :raises AttributeError: If the connection is not initialized or the transaction is not started.
        """
        await self.app.sqlalchemy.end_transaction(self.transaction, rollback=rollback)
        del self.transaction

        await self.app.sqlalchemy.close_connection(self.connection)
        del self.connection

    async def repository_params(self) -> tuple[list[t.Any], dict[str, t.Any]]:
        """Get the parameters for initialising the repositories.

        :return: Parameters for initialising the repositories.
        """
        return [self.connection], {}

    async def commit(self):
        """Commit the unit of work."""
        await self.connection.commit()

    async def rollback(self):
        """Rollback the unit of work."""
        await self.connection.rollback()



import dataclasses
from unittest.mock import MagicMock, patch

import pytest

from flama import http
from flama.applications import Flama
from flama.debug.data_structures import App, Environment, Error, ErrorContext, NotFoundContext, Request
from flama.routing import Mount, Route


@pytest.fixture
def http_request():
    return MagicMock()


class TestCaseRequest:
    def test_from_request(self, asgi_scope, asgi_receive, asgi_send):
        request = http.Request(asgi_scope, asgi_receive, asgi_send)

        result = dataclasses.asdict(Request.from_request(request))

        assert result == {
            "client": None,
            "cookies": {},
            "headers": {},
            "method": "GET",
            "params": {"path": {}, "query": {}},
            "path": "/",
        }


class TestCaseError:
    def test_from_exception(self):
        try:
            raise ValueError("Foo")
        except ValueError as e:
            result = dataclasses.asdict(Error.from_exception(e))

        traceback = result.pop("traceback", None)
        assert traceback
        assert len(traceback) == 1
        frame = traceback[0]
        code = frame.pop("code", None)
        assert code
        assert frame == {
            "filename": "tests/debug/test_data_structures.py",
            "function": "test_from_exception",
            "line": 36,
            "vendor": None,
        }
        assert result == {"description": "Foo", "error": "ValueError"}


class TestCaseEnvironment:
    def test_from_system(self):
        result = dataclasses.asdict(Environment.from_system())

        path = result.pop("path", None)
        assert path
        assert list(result.keys()) == ["platform", "python", "python_version"]


class TestCaseApp:
    def test_from_app(self):
        def foo_handler(): ...

        def bar_handler(): ...

        app = Flama(
            routes=[
                Route("/", foo_handler, name="foo"),
                Mount("/subapp/", routes=[Route("/", bar_handler, name="bar")], name="subapp"),
            ],
            schema=None,
            docs=None,
        )

        result = dataclasses.asdict(App.from_app(app))

        assert result == {
            "name": None,
            "path": "/",
            "endpoints": [
                {
                    "path": "/",
                    "endpoint": "foo_handler",
                    "module": "tests.debug.test_data_structures",
                    "file": "tests/debug/test_data_structures.py",
                    "line": 66,
                    "name": "foo",
                },
            ],
            "apps": [
                {
                    "name": "subapp",
                    "path": "/subapp/",
                    "apps": [],
                    "endpoints": [
                        {
                            "path": "/",
                            "module": "tests.debug.test_data_structures",
                            "file": "tests/debug/test_data_structures.py",
                            "line": 68,
                            "endpoint": "bar_handler",
                            "name": "bar",
                        }
                    ],
                },
            ],
        }


class TestCaseErrorContext:
    def test_build(self):
        request_mock = MagicMock(Request)
        environment_mock = MagicMock(Environment)
        error_mock = MagicMock(Error)

        with (
            patch.object(Request, "from_request", return_value=request_mock),
            patch.object(Environment, "from_system", return_value=environment_mock),
            patch.object(Error, "from_exception", return_value=error_mock),
        ):
            context = dataclasses.asdict(ErrorContext.build(MagicMock(), MagicMock()))

        assert context == {"request": request_mock, "environment": environment_mock, "error": error_mock}


class TestCaseNotFoundContext:
    def test_build(self):
        request_mock = MagicMock(Request)
        environment_mock = MagicMock(Environment)
        app_mock = MagicMock(App)

        with (
            patch.object(Request, "from_request", return_value=request_mock),
            patch.object(Environment, "from_system", return_value=environment_mock),
            patch.object(App, "from_app", return_value=app_mock),
        ):
            context = dataclasses.asdict(NotFoundContext.build(MagicMock(), MagicMock()))

        assert context == {"request": request_mock, "environment": environment_mock, "app": app_mock}



import os
import typing as t

from flama.serialize.data_structures import ModelArtifact

__all__ = ["load"]


def load(path: t.Union[str, os.PathLike], **kwargs) -> ModelArtifact:
    """Deserialize a ML model using Flama format from a bytes stream.

    :param path: Model file path.
    :param kwargs: Keyword arguments passed to library load method.
    :return: Model artifact.
    """
    return ModelArtifact.load(path, **kwargs)



from unittest.mock import PropertyMock, patch

import marshmallow
import pydantic
import pytest
import typesystem

from flama import Flama, exceptions


class TestCaseSetup:
    def test_setup_default(self):
        Flama()

        from flama import schemas

        assert schemas.lib == pydantic

    def test_setup_pydantic(self):
        Flama(schema_library="pydantic")

        from flama import schemas

        assert schemas.lib == pydantic

    def test_setup_typesystem(self):
        Flama(schema_library="typesystem")

        from flama import schemas

        assert schemas.lib == typesystem

    def test_setup_marshmallow(self):
        Flama(schema_library="marshmallow")

        from flama import schemas

        assert schemas.lib == marshmallow

    def test_setup_no_lib_installed(self):
        from flama import schemas

        with (
            patch("flama.schemas.Module.available", PropertyMock(return_value=iter(()))),
            pytest.raises(
                exceptions.DependencyNotInstalled,
                match="No schema library is installed. Install one of your preference following instructions from: "
                "https://flama.dev/docs/getting-started/installation#extras",
            ),
        ):
            schemas._module.setup()



import typing as t
from unittest.mock import AsyncMock, MagicMock, call, patch

import marshmallow
import pydantic
import pytest
import typesystem
import typesystem.fields

from flama import Component, Flama, endpoints, schemas, types


class Puppy:
    name = "Canna"


class PuppyComponent(Component):
    def resolve(self) -> Puppy:
        return Puppy()


class TestCaseHTTPEndpoint:
    @pytest.fixture(scope="class")
    def app(self, app):
        return Flama(schema=None, docs=None, components=[PuppyComponent()])

    @pytest.fixture
    def endpoint(self, app, asgi_scope, asgi_receive, asgi_send):
        @app.route("/")
        class FooEndpoint(endpoints.HTTPEndpoint):
            def get(self): ...

        asgi_scope["app"] = app
        asgi_scope["root_app"] = app
        asgi_scope["type"] = "http"
        return FooEndpoint(asgi_scope, asgi_receive, asgi_send)

    @pytest.fixture(scope="class")
    def puppy_schema(self, app):
        if app.schema.schema_library.lib == pydantic:
            schema = pydantic.create_model("Puppy", name=(str, ...))
        elif app.schema.schema_library.lib == typesystem:
            schema = typesystem.Schema(title="Puppy", fields={"name": typesystem.fields.String()})
        elif app.schema.schema_library.lib == marshmallow:
            schema = type("Puppy", (marshmallow.Schema,), {"name": marshmallow.fields.String()})
        else:
            raise ValueError("Wrong schema lib")

        return schema

    @pytest.fixture(scope="class")
    def puppy_endpoint(self, app, puppy_schema):
        @app.route("/puppy/")
        class PuppyEndpoint(endpoints.HTTPEndpoint):
            def get(self, puppy: Puppy) -> t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(puppy_schema)]:
                return {"name": puppy.name}

            async def post(
                self, puppy: t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(puppy_schema)]
            ) -> t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(puppy_schema)]:
                return puppy

        return PuppyEndpoint

    @pytest.mark.parametrize(
        ["method", "params", "status_code", "expected_response"],
        (
            pytest.param("get", {}, 200, {"name": "Canna"}, id="get"),
            pytest.param("post", {"json": {"name": "Canna"}}, 200, {"name": "Canna"}, id="post"),
            pytest.param(
                "patch",
                {},
                405,
                {"detail": "Method Not Allowed", "error": "HTTPException", "status_code": 405},
                id="method_not_allowed",
            ),
        ),
    )
    async def test_request(self, app, client, puppy_endpoint, method, params, status_code, expected_response):
        response = await client.request(method, "/puppy/", **params)

        assert response.status_code == status_code
        assert response.json() == expected_response

    def test_init(self, app, asgi_scope, asgi_receive, asgi_send):
        with patch("flama.http.Request") as request_mock:

            class FooEndpoint(endpoints.HTTPEndpoint):
                def get(self): ...

            route = app.add_route("/", FooEndpoint)
            asgi_scope = types.Scope(
                {
                    **asgi_scope,
                    "app": app,
                    "root_app": app,
                    "type": "http",
                    "method": "GET",
                    "path": "/",
                    "path_params": {},
                    "endpoint": FooEndpoint,
                    "route": route,
                }
            )
            endpoint = endpoints.HTTPEndpoint(asgi_scope, asgi_receive, asgi_send)
            assert endpoint.state == {
                "scope": asgi_scope,
                "receive": asgi_receive,
                "send": asgi_send,
                "exc": None,
                "app": app,
                "root_app": app,
                "path_params": {},
                "route": route,
                "request": request_mock(),
            }

    def test_allowed_methods(self, endpoint):
        assert endpoint.allowed_methods() == {"GET", "HEAD"}

        class BarEndpoint(endpoints.HTTPEndpoint):
            def post(self): ...

        assert BarEndpoint.allowed_methods() == {"POST"}

    def test_handler(self, endpoint):
        endpoint.state["request"].scope["method"] = "GET"
        assert endpoint.handler == endpoint.get

    async def test_dispatch(self, app, endpoint):
        injected_mock = MagicMock()
        app.injector.inject = AsyncMock(return_value=injected_mock)
        with patch("flama.concurrency.run") as run_mock:
            await endpoint.dispatch()

            assert app.injector.inject.call_args_list == [call(endpoint.get, endpoint.state)]
            assert run_mock.call_args_list == [call(injected_mock)]



__all__ = ["assert_recursive_contains"]


def assert_recursive_contains(expected, value):
    if isinstance(expected, dict) and isinstance(value, dict):
        assert expected.keys() <= value.keys()

        if expected != value:
            for k, v in expected.items():
                assert_recursive_contains(v, value[k])
    elif isinstance(expected, (list, set, tuple)) and isinstance(value, (list, set, tuple)):
        assert len(expected) <= len(value)

        if expected != value:
            for i, _ in enumerate(expected):
                assert_recursive_contains(expected[i], value[i])
    else:
        assert expected == value



import dataclasses
import datetime
import enum
import json
import pathlib
import uuid
from unittest.mock import AsyncMock, MagicMock, Mock, call, mock_open, patch

import jinja2
import marshmallow
import pydantic
import pytest
import typesystem
import typesystem.fields

from flama import exceptions, http, types


@dataclasses.dataclass
class Foo:
    bar: int


class TestCaseRequest:
    @pytest.fixture(scope="function", autouse=True)
    def add_endpoints(self, app):
        @app.route("/request/")
        async def get_request(request: http.Request):
            return {
                "method": str(request.method),
                "url": str(request.url),
                "headers": dict(request.headers),
                "body": (await request.body()).decode("utf-8"),
            }

    async def test_request(self, client):
        expected_response = {
            "method": "GET",
            "url": "http://localapp/request/",
            "headers": {
                "accept": "*/*",
                "accept-encoding": "gzip, deflate",
                "connection": "keep-alive",
                "host": "localapp",
            },
            "body": "",
        }

        response = await client.get("/request/")
        response_json = response.json()
        del response_json["headers"]["user-agent"]

        assert response_json == expected_response, str(response_json)


class TestCaseResponse:
    async def test_call(self):
        response = http.Response()
        scope, receive, send = types.Scope({}), AsyncMock(), AsyncMock()

        with patch("starlette.responses.Response.__call__") as call_mock:
            await response(scope, receive, send)

            assert call_mock.call_args_list == [call(scope, receive, send)]

    def test_eq(self):
        assert http.Response(content="foo") == http.Response(content="foo")
        assert http.Response(content="foo") != http.Response(content="bar")


class TestCaseHTMLResponse:
    async def test_call(self):
        response = http.HTMLResponse()
        scope, receive, send = types.Scope({}), AsyncMock(), AsyncMock()

        with patch("starlette.responses.HTMLResponse.__call__") as call_mock:
            await response(scope, receive, send)

            assert call_mock.call_args_list == [call(scope, receive, send)]


class TestCasePlainTextResponse:
    async def test_call(self):
        response = http.PlainTextResponse()
        scope, receive, send = types.Scope({}), AsyncMock(), AsyncMock()

        with patch("starlette.responses.PlainTextResponse.__call__") as call_mock:
            await response(scope, receive, send)

            assert call_mock.call_args_list == [call(scope, receive, send)]


class TestCaseJSONResponse:
    @pytest.fixture
    def schema(self):
        return Mock()

    async def test_call(self):
        response = http.JSONResponse(content={})
        scope, receive, send = types.Scope({}), AsyncMock(), AsyncMock()

        with patch("starlette.responses.JSONResponse.__call__") as call_mock:
            await response(scope, receive, send)

            assert call_mock.call_args_list == [call(scope, receive, send)]

    @pytest.mark.parametrize(
        ["content", "result", "exception"],
        (
            pytest.param(
                {"foo": {"bar": [1, "foobar", 2.0, True, None]}},
                {"foo": {"bar": [1, "foobar", 2.0, True, None]}},
                None,
                id="default",
            ),
            pytest.param({"foo": pathlib.Path("foo/bar.json")}, {"foo": "foo/bar.json"}, None, id="path"),
            pytest.param({"foo": b"bar"}, {"foo": "bar"}, None, id="bytes"),
            pytest.param({"foo": bytearray([1, 2, 3])}, {"foo": "\x01\x02\x03"}, None, id="bytearray"),
            pytest.param({"foo": enum.Enum("Foo", ["bar"]).bar}, {"foo": 1}, None, id="enum"),
            pytest.param({"foo": uuid.UUID(int=0)}, {"foo": "00000000-0000-0000-0000-000000000000"}, None, id="uuid"),
            pytest.param({"foo": {"bar"}}, {"foo": ["bar"]}, None, id="set"),
            pytest.param({"foo": frozenset({"bar"})}, {"foo": ["bar"]}, None, id="set"),
            pytest.param(
                {"foo": datetime.datetime(2023, 9, 20, 11, 30, 0)}, {"foo": "2023-09-20T11:30:00"}, None, id="datetime"
            ),
            pytest.param({"foo": datetime.date(2023, 9, 20)}, {"foo": "2023-09-20"}, None, id="date"),
            pytest.param({"foo": datetime.time(11, 30, 0)}, {"foo": "11:30:00"}, None, id="time"),
            pytest.param(
                {"foo": datetime.timedelta(days=1, hours=20, minutes=30, seconds=10, milliseconds=10, microseconds=6)},
                {"foo": "P1D20H30M10.010006S"},
                None,
                id="timedelta",
            ),
            pytest.param({"foo": Exception}, {"foo": "Exception"}, None, id="exception_class"),
            pytest.param({"foo": Exception("bar")}, {"foo": "Exception('bar')"}, None, id="exception_obj"),
            pytest.param({"foo": "bar"}, {"foo": "bar"}, None, id="schema"),
            pytest.param({"foo": Foo(bar=1)}, {"foo": {"bar": 1}}, None, id="dataclass"),
            pytest.param({"foo": Mock()}, None, TypeError, id="error"),
        ),
        indirect=["exception"],
    )
    def test_render(self, content, result, exception):
        with exception:
            response = http.JSONResponse(content=content)

            assert json.loads(response.body.decode()) == result


class TestCaseRedirectResponse:
    async def test_call(self):
        response = http.RedirectResponse(url="")
        scope, receive, send = types.Scope({}), AsyncMock(), AsyncMock()

        with patch("starlette.responses.RedirectResponse.__call__") as call_mock:
            await response(scope, receive, send)

            assert call_mock.call_args_list == [call(scope, receive, send)]


class TestCaseStreamingResponse:
    async def test_call(self):
        response = http.StreamingResponse(content=(x for x in "foo"))
        scope, receive, send = types.Scope({}), AsyncMock(), AsyncMock()

        with patch("starlette.responses.StreamingResponse.__call__") as call_mock:
            await response(scope, receive, send)

            assert call_mock.call_args_list == [call(scope, receive, send)]


class TestCaseFileResponse:
    async def test_call(self):
        response = http.FileResponse(path="")
        scope, receive, send = types.Scope({}), AsyncMock(), AsyncMock()

        with patch("starlette.responses.FileResponse.__call__") as call_mock:
            await response(scope, receive, send)

            assert call_mock.call_args_list == [call(scope, receive, send)]


class TestCaseAPIResponse:
    @pytest.fixture(scope="function")
    def schema(self, app):
        if app.schema.schema_library.lib == pydantic:
            schema = pydantic.create_model("Puppy", name=(str, ...))
        elif app.schema.schema_library.lib == typesystem:
            schema = typesystem.Schema(title="Puppy", fields={"name": typesystem.fields.String()})
        elif app.schema.schema_library.lib == marshmallow:
            schema = type("Puppy", (marshmallow.Schema,), {"name": marshmallow.fields.String(required=True)})
        else:
            raise ValueError("Wrong schema lib")

        return schema

    def test_init(self, schema):
        with patch("flama.http.starlette.responses.JSONResponse.__init__"):
            response = http.APIResponse(schema=schema)

        assert response.schema == schema

    @pytest.mark.parametrize(
        ["use_schema", "content", "expected", "exception"],
        (
            pytest.param(True, {"name": "Canna"}, '{"name":"Canna"}', None, id="schema_and_content"),
            pytest.param(False, {}, "", None, id="no_content"),
            pytest.param(False, {"name": "Canna"}, '{"name":"Canna"}', None, id="no_schema"),
            pytest.param(True, {"foo": "bar"}, "", exceptions.SerializationError, id="error"),
        ),
        indirect=["exception"],
    )
    def test_render(self, schema, use_schema, content, expected, exception):
        with exception:
            response = http.APIResponse(schema=schema if use_schema else None, content=content)
            assert response.body.decode() == expected


class TestCaseAPIErrorResponse:
    def test_init(self):
        detail = "foo"
        exception = ValueError()
        status_code = 401
        expected_result = {"detail": "foo", "error": "ValueError", "status_code": 401}

        response = http.APIErrorResponse(detail=detail, status_code=status_code, exception=exception)

        assert response.detail == detail
        assert response.exception == exception
        assert response.status_code == status_code
        assert json.loads(response.body.decode()) == expected_result


class TestCaseHTMLFileResponse:
    def test_init(self):
        content = "<html></html>"
        with patch("builtins.open", mock_open(read_data=content)):
            response = http.HTMLFileResponse("foo.html")

        assert response.body == content.encode()

    def test_init_error(self):
        error_detail = "Foo error"
        with (
            patch("builtins.open", side_effect=ValueError(error_detail)),
            pytest.raises(exceptions.HTTPException) as exc,
        ):
            http.HTMLFileResponse("foo.html")

            assert exc.status_code == 500
            assert exc.detail == error_detail


class TestCaseHTMLTemplatesEnvironment:
    @pytest.fixture
    def environment(self):
        return http.HTMLTemplatesEnvironment()

    @pytest.mark.parametrize(
        ["value", "result"],
        (
            pytest.param("& < > \" ' \n", "&amp; &lt; &gt; &quot; &#x27; &#10;&#13;", id="str"),
            pytest.param(1, 1, id="int"),
            pytest.param(1.0, 1.0, id="float"),
            pytest.param(True, True, id="bool"),
            pytest.param(None, None, id="none"),
            pytest.param(
                ["foo", 1, 1.0, True, None, ["nested", "&"], ("nested", "&"), {"nested": "&"}],
                ["foo", 1, 1.0, True, None, ["nested", "&amp;"], ["nested", "&amp;"], {"nested": "&amp;"}],
                id="list",
            ),
            pytest.param(
                ("foo", 1, 1.0, True, None, ["nested", "&"], ("nested", "&"), {"nested": "&"}),
                ["foo", 1, 1.0, True, None, ["nested", "&amp;"], ["nested", "&amp;"], {"nested": "&amp;"}],
                id="tuple",
            ),
            pytest.param(
                {
                    "str": "foo",
                    "int": 1,
                    "float": 1.0,
                    "bool": True,
                    "null": None,
                    "list": ["nested", "&"],
                    "tuple": ("nested", "&"),
                    "dict": {"nested": "&"},
                },
                {
                    "str": "foo",
                    "int": 1,
                    "float": 1.0,
                    "bool": True,
                    "null": None,
                    "list": ["nested", "&amp;"],
                    "tuple": ["nested", "&amp;"],
                    "dict": {"nested": "&amp;"},
                },
                id="dict",
            ),
        ),
    )
    def test_escape(self, environment, value, result):
        assert environment._escape(value) == result

    @pytest.mark.parametrize(
        ["value", "result"],
        (
            pytest.param("& < > \" ' \n", "&amp; &lt; &gt; &quot; &#x27; &#10;&#13;", id="str"),
            pytest.param("1", "1", id="int"),
            pytest.param("1.0", "1.0", id="float"),
            pytest.param("true", "true", id="bool"),
            pytest.param("null", "null", id="none"),
        ),
    )
    def test_safe(self, environment, value, result):
        assert environment.safe(value) == result

    @pytest.mark.parametrize(
        ["value", "result"],
        (
            pytest.param("& < > \" ' \n", '\\"&amp; &lt; &gt; &quot; &#x27; &#10;&#13;\\"', id="str"),
            pytest.param(1, "1", id="int"),
            pytest.param(1.0, "1.0", id="float"),
            pytest.param(True, "true", id="bool"),
            pytest.param(None, "null", id="none"),
            pytest.param(
                ["foo", 1, 1.0, True, None, ["nested", "&"], ("nested", "&"), {"nested": "&"}],
                '[\\"foo\\", 1, 1.0, true, null, [\\"nested\\", \\"&amp;\\"], [\\"nested\\", \\"&amp;\\"], '
                '{\\"nested\\": \\"&amp;\\"}]',
                id="list",
            ),
            pytest.param(
                ("foo", 1, 1.0, True, None, ["nested", "&"], ("nested", "&"), {"nested": "&"}),
                '[\\"foo\\", 1, 1.0, true, null, [\\"nested\\", \\"&amp;\\"], [\\"nested\\", \\"&amp;\\"], '
                '{\\"nested\\": \\"&amp;\\"}]',
                id="tuple",
            ),
            pytest.param(
                {
                    "str": "foo",
                    "int": 1,
                    "float": 1.0,
                    "bool": True,
                    "null": None,
                    "list": ["nested", "&"],
                    "tuple": ("nested", "&"),
                    "dict": {"nested": "&"},
                },
                '{\\"str\\": \\"foo\\", \\"int\\": 1, \\"float\\": 1.0, \\"bool\\": true, \\"null\\": null, '
                '\\"list\\": [\\"nested\\", \\"&amp;\\"], \\"tuple\\": [\\"nested\\", \\"&amp;\\"], '
                '\\"dict\\": {\\"nested\\": \\"&amp;\\"}}',
                id="dict",
            ),
        ),
    )
    def test_safe_json(self, environment, value, result):
        assert environment.safe_json(value) == result


class TestCaseHTMLTemplateResponse:
    @pytest.mark.parametrize(
        ["context"], (pytest.param({"foo": "bar"}, id="context"), pytest.param(None, id="no_context"))
    )
    def test_init(self, context):
        template_mock = MagicMock()
        template_mock.render.return_value = "foo"
        environment_mock = MagicMock(spec=jinja2.Environment)
        environment_mock.get_template.return_value = template_mock
        with (
            patch.object(http.HTMLTemplateResponse, "templates", new=environment_mock),
            patch.object(http.HTMLResponse, "__init__", return_value=None) as super_mock,
        ):
            http.HTMLTemplateResponse("foo.html", context)

            assert super_mock.call_args_list == [call("foo")]


class TestCaseFlamaLoader:
    @pytest.mark.parametrize(
        ["import_mock", "path_exists_mock", "mkdir_call", "exception"],
        [
            pytest.param(
                MagicMock(find_spec=MagicMock()),
                True,
                False,
                None,
                id="ok",
            ),
            pytest.param(
                MagicMock(find_spec=MagicMock()),
                False,
                True,
                None,
                id="ok_create_templates_path",
            ),
            pytest.param(
                MagicMock(find_spec=MagicMock(return_value=None)),
                True,
                False,
                exceptions.ApplicationError("Flama package not found"),
                id="error_flama_package_not_found",
            ),
        ],
        indirect=["exception"],
    )
    def test_init(self, import_mock, path_exists_mock, mkdir_call, exception):
        with (
            exception,
            patch("jinja2.PackageLoader.__init__"),
            patch("importlib.util", import_mock),
            patch.object(pathlib.Path, "exists", return_value=path_exists_mock),
            patch.object(pathlib.Path, "mkdir") as mkdir_mock,
        ):
            http._FlamaLoader()

            if mkdir_call:
                assert mkdir_mock.call_args_list == [call(exist_ok=True)]


class TestCaseOpenAPIResponse:
    @pytest.mark.parametrize(
        "test_input,expected,exception",
        (
            pytest.param({"foo": "bar"}, {"foo": "bar"}, None, id="success"),
            pytest.param("foo", None, ValueError("The schema must be a dictionary"), id="wrong_content"),
        ),
        indirect=("exception",),
    )
    def test_render(self, test_input, expected, exception):
        with exception:
            response = http.OpenAPIResponse(test_input)

            assert json.loads(response.body.decode()) == expected

    async def test_call(self):
        response = http.OpenAPIResponse(content={})
        scope, receive, send = types.Scope({}), AsyncMock(), AsyncMock()

        with patch("starlette.schemas.OpenAPIResponse.__call__") as call_mock:
            await response(scope, receive, send)

            assert call_mock.call_args_list == [call(scope, receive, send)]



import sys

__all__ = ["Concatenate", "ParamSpec", "TypeGuard", "UnionType", "NotRequired", "StrEnum", "tomllib"]

# PORT: Remove when stop supporting 3.9
# Concatenate was added in Python 3.10
# https://docs.python.org/3/library/typing.html#typing.Concatenate
if sys.version_info >= (3, 10):
    from typing import Concatenate
else:
    from typing_extensions import Concatenate


# PORT: Remove when stop supporting 3.9
# ParamSpec was added in Python 3.10
# https://docs.python.org/3/library/typing.html#typing.ParamSpec
if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec


# PORT: Remove when stop supporting 3.9
# TypeGuard was added in Python 3.10
# https://docs.python.org/3/library/typing.html#typing.TypeGuard
if sys.version_info >= (3, 10):
    from typing import TypeGuard
else:
    from typing_extensions import TypeGuard


# PORT: Remove when stop supporting 3.9
# UnionType was added in Python 3.10
# https://docs.python.org/3/library/stdtypes.html#types-union
if sys.version_info >= (3, 10):
    from types import UnionType
else:
    from typing import Union as UnionType

# PORT: Remove when stop supporting 3.10
# NotRequired was added in Python 3.11
# https://docs.python.org/3/library/enum.html#enum.StrEnum
if sys.version_info >= (3, 11):
    from typing import NotRequired
else:
    from typing_extensions import NotRequired


# PORT: Remove when stop supporting 3.10
# StrEnum was added in Python 3.11
# https://docs.python.org/3/library/enum.html#enum.StrEnum
if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from enum import Enum

    class StrEnum(str, Enum):
        @staticmethod
        def _generate_next_value_(name, start, count, last_values):
            return name.lower()


# PORT: Remove when stop supporting 3.10
# Tomllib was added in Python 3.11
# https://docs.python.org/3/library/tomllib.html
if sys.version_info >= (3, 11):  # PORT: Remove when stop supporting 3.10
    import tomllib
else:
    try:
        import tomli

        tomllib = tomli
    except ModuleNotFoundError:
        tomllib = None



import marshmallow

__all__ = [
    "APIError",
    "DropCollection",
    "LimitOffsetMeta",
    "LimitOffset",
    "PageNumberMeta",
    "PageNumber",
    "MLModelInput",
    "MLModelOutput",
    "SCHEMAS",
]


class APIError(marshmallow.Schema):
    status_code = marshmallow.fields.Integer(
        metadata={"title": "status_code", "description": "HTTP status code"}, required=True
    )
    detail = marshmallow.fields.Raw(metadata={"title": "detail", "description": "Error detail"}, required=True)
    error = marshmallow.fields.String(metadata={"title": "type", "description": "Exception or error type"})


class DropCollection(marshmallow.Schema):
    deleted = marshmallow.fields.Integer(
        metadata={"title": "deleted", "description": "Number of deleted elements"}, required=True
    )


class LimitOffsetMeta(marshmallow.Schema):
    limit = marshmallow.fields.Integer(metadata={"title": "limit", "description": "Number of retrieved items"})
    offset = marshmallow.fields.Integer(metadata={"title": "offset", "description": "Collection offset"})
    count = marshmallow.fields.Integer(
        metadata={"title": "count", "description": "Total number of items"}, allow_none=True
    )


class LimitOffset(marshmallow.Schema):
    meta = marshmallow.fields.Nested(
        LimitOffsetMeta(), required=True, metadata={"title": "meta", "description": "Pagination metadata"}
    )
    data = marshmallow.fields.List(
        marshmallow.fields.Dict(), required=True, metadata={"title": "data", "description": "Paginated data"}
    )


class PageNumberMeta(marshmallow.Schema):
    page = marshmallow.fields.Integer(metadata={"title": "page", "description": "Current page number"})
    page_size = marshmallow.fields.Integer(metadata={"title": "page_size", "description": "Page size"})
    count = marshmallow.fields.Integer(
        metadata={"title": "count", "description": "Total number of items"}, allow_none=True
    )


class PageNumber(marshmallow.Schema):
    meta = marshmallow.fields.Nested(
        PageNumberMeta(), required=True, metadata={"title": "meta", "description": "Pagination metadata"}
    )
    data = marshmallow.fields.List(
        marshmallow.fields.Dict(), required=True, metadata={"title": "data", "description": "Paginated data"}
    )


class MLModelInput(marshmallow.Schema):
    input = marshmallow.fields.List(
        marshmallow.fields.Raw(),
        required=True,
        metadata={"title": "input", "description": "Model input"},
    )


class MLModelOutput(marshmallow.Schema):
    output = marshmallow.fields.List(
        marshmallow.fields.Raw(),
        required=True,
        metadata={"title": "output", "description": "Model output"},
    )


SCHEMAS = {
    "flama.APIError": APIError,
    "flama.DropCollection": DropCollection,
    "flama.LimitOffsetMeta": LimitOffsetMeta,
    "flama.LimitOffset": LimitOffset,
    "flama.PageNumberMeta": PageNumberMeta,
    "flama.PageNumber": PageNumber,
    "flama.MLModelInput": MLModelInput,
    "flama.MLModelOutput": MLModelOutput,
}



import sqlalchemy
from pydantic import BaseModel

import flama
from flama import Flama
from flama.resources.crud import CRUDResource
from flama.sqlalchemy import SQLAlchemyModule

DATABASE_URL = "sqlite+aiosqlite://"

metadata = sqlalchemy.MetaData()


class PuppySchema(BaseModel):
    custom_id: int
    name: str


PuppyModel = sqlalchemy.Table(
    "puppy",
    metadata,
    sqlalchemy.Column("custom_id", sqlalchemy.Integer, primary_key=True, autoincrement=True),
    sqlalchemy.Column("name", sqlalchemy.String),
)


class PuppyResource(CRUDResource):
    name = "puppy"
    verbose_name = "Puppy"

    model = PuppyModel
    schema = PuppySchema


app = Flama(
    openapi={
        "info": {
            "title": "Puppy Register",  # API title
            "version": "0.1.0",  # API version
            "description": "A register of puppies",  # API description
        }
    },
    modules=[SQLAlchemyModule(database=DATABASE_URL)],
)

app.resources.add_resource("/", PuppyResource)


if __name__ == "__main__":
    flama.run(flama_app=app, server_host="0.0.0.0", server_port=8080)



import codecs
import dataclasses
import datetime
import enum
import importlib
import inspect
import json
import logging
import os
import shutil
import tarfile
import tempfile
import typing as t
import uuid
import warnings
import weakref
from pathlib import Path

from flama import compat, exceptions
from flama.serialize.base import Serializer
from flama.serialize.types import Framework

if t.TYPE_CHECKING:
    from flama.types import JSONSchema

__all__ = ["ModelArtifact", "Compression"]

logger = logging.getLogger(__name__)


class Compression(compat.StrEnum):  # PORT: Replace compat when stop supporting 3.10
    fast = enum.auto()
    standard = enum.auto()
    high = enum.auto()

    @property
    def compression_format(self) -> str:
        return {
            Compression.fast: "gz",
            Compression.standard: "bz2",
            Compression.high: "xz",
        }[self]


class FrameworkSerializers:
    @classmethod
    def serializer(cls, framework: t.Union[str, Framework]) -> Serializer:
        try:
            module, class_name = {
                Framework.torch: ("pytorch", "PyTorchSerializer"),
                Framework.sklearn: ("sklearn", "SKLearnSerializer"),
                Framework.keras: ("tensorflow", "TensorFlowSerializer"),
                Framework.tensorflow: ("tensorflow", "TensorFlowSerializer"),
            }[Framework[framework]]
        except KeyError:  # pragma: no cover
            raise ValueError("Wrong framework")

        serializer_class: type[Serializer] = getattr(
            importlib.import_module(f"flama.serialize.serializers.{module}"), class_name
        )
        return serializer_class()

    @classmethod
    def from_model(cls, model: t.Any) -> Serializer:
        inspect_objs = [model]

        try:
            inspect_objs += model.__class__.__mro__
        except AttributeError:  # pragma: no cover
            ...

        for obj in inspect_objs:
            try:
                return cls.serializer(inspect.getmodule(obj).__name__.split(".", 1)[0])  # type: ignore[union-attr]
            except (ValueError, AttributeError):
                ...
        else:
            raise ValueError("Unknown model framework")


@dataclasses.dataclass
class FrameworkInfo:
    """Dataclass for storing model framework information."""

    lib: Framework
    version: str

    @classmethod
    def from_model(cls, model: t.Any) -> "FrameworkInfo":
        serializer = FrameworkSerializers.from_model(model)
        return cls(lib=serializer.lib, version=serializer.version())

    @classmethod
    def from_dict(cls, data: dict[str, t.Any]) -> "FrameworkInfo":
        return cls(lib=Framework[data["lib"]], version=data["version"])

    def to_dict(self) -> dict[str, t.Any]:
        return {"lib": self.lib.value, "version": self.version}


@dataclasses.dataclass
class ModelInfo:
    """Dataclass for storing model info."""

    obj: str
    info: t.Optional["JSONSchema"] = None
    params: t.Optional[dict[str, t.Any]] = None
    metrics: t.Optional[dict[str, t.Any]] = None

    @classmethod
    def from_model(
        cls, model: t.Any, params: t.Optional[dict[str, t.Any]], metrics: t.Optional[dict[str, t.Any]]
    ) -> "ModelInfo":
        return cls(
            obj=model.__name__ if inspect.isclass(model) else model.__class__.__name__,
            info=FrameworkSerializers.from_model(model).info(model),
            params=params,
            metrics=metrics,
        )

    @classmethod
    def from_dict(cls, data: dict[str, t.Any]):
        return cls(obj=data["obj"], info=data["info"], params=data.get("params"), metrics=data.get("metrics"))

    def to_dict(self) -> dict[str, t.Any]:
        return dataclasses.asdict(self)


@dataclasses.dataclass(frozen=True)
class Metadata:
    """Dataclass for storing model metadata."""

    id: t.Union[str, uuid.UUID]
    timestamp: datetime.datetime
    framework: FrameworkInfo
    model: ModelInfo
    extra: t.Optional[dict[str, t.Any]] = None

    @classmethod
    def from_model(
        cls,
        model: t.Any,
        *,
        model_id: t.Optional[t.Union[str, uuid.UUID]],
        timestamp: t.Optional[datetime.datetime],
        params: t.Optional[dict[str, t.Any]],
        metrics: t.Optional[dict[str, t.Any]],
        extra: t.Optional[dict[str, t.Any]],
    ) -> "Metadata":
        return cls(
            id=model_id or uuid.uuid4(),
            timestamp=timestamp or datetime.datetime.now(),
            framework=FrameworkInfo.from_model(model),
            model=ModelInfo.from_model(model, params, metrics),
            extra=extra,
        )

    @classmethod
    def from_dict(cls, data: dict[str, t.Any]) -> "Metadata":
        try:
            id_ = uuid.UUID(data["id"])
        except ValueError:
            id_ = data["id"]

        timestamp = (
            datetime.datetime.fromisoformat(data["timestamp"])
            if isinstance(data["timestamp"], str)
            else data["timestamp"]
        )

        return cls(
            id=id_,
            timestamp=timestamp,
            framework=FrameworkInfo.from_dict(data["framework"]),
            model=ModelInfo.from_dict(data["model"]),
            extra=data.get("extra"),
        )

    def to_dict(self) -> dict[str, t.Any]:
        return {
            "id": str(self.id),
            "timestamp": self.timestamp.isoformat(),
            "framework": self.framework.to_dict(),
            "model": self.model.to_dict(),
            "extra": self.extra,
        }


Artifacts = dict[str, t.Union[str, os.PathLike]]


class _ModelDirectory:
    def __init__(
        self,
        model_file: t.Union[str, os.PathLike],
        path: t.Optional[t.Union[str, os.PathLike]] = None,
        delete: bool = True,
    ):
        """Generate a model directory from a model file.

        :param model_file: Model file path.
        :param path: Directory path. Create a temporary directory if None.
        :return: Model directory loaded.
        """
        self.directory = Path(path) if path else Path(tempfile.mkdtemp())

        with tarfile.open(model_file, "r") as tar:
            tar.extractall(self.directory)

        self.model = self.directory / "model"
        self.artifacts: Artifacts = {artifact.name: artifact for artifact in self.directory.glob("artifacts/*")}

        logger.debug("Model '%s' extracted in directory '%s'", model_file, self.directory)

        self._finalizer = weakref.finalize(self, self._cleanup) if delete else None

    def _cleanup(self):
        logger.debug("Model directory '%s' clean", self.directory)
        shutil.rmtree(self.directory)

    def exists(self) -> bool:
        """Check if the model directory exists.

        :return: True if the directory exists.
        """
        return os.path.exists(self.directory)

    def cleanup(self):
        """Clean the model directory by removing it."""
        if (self._finalizer and self._finalizer.detach()) or self.exists():
            self._cleanup()


@dataclasses.dataclass(frozen=True)
class ModelArtifact:
    """ML Model wrapper to provide mechanisms for serialization and deserialization using Flama format."""

    model: t.Any
    meta: Metadata
    artifacts: t.Optional[Artifacts] = None
    _directory: t.Optional[_ModelDirectory] = dataclasses.field(default=None, repr=False, compare=False)

    @classmethod
    def from_model(
        cls,
        model: t.Any,
        *,
        model_id: t.Optional[t.Union[str, uuid.UUID]] = None,
        timestamp: t.Optional[datetime.datetime] = None,
        params: t.Optional[dict[str, t.Any]] = None,
        metrics: t.Optional[dict[str, t.Any]] = None,
        extra: t.Optional[dict[str, t.Any]] = None,
        artifacts: t.Optional[Artifacts] = None,
    ) -> "ModelArtifact":
        return cls(
            model=model,
            meta=Metadata.from_model(
                model, model_id=model_id, timestamp=timestamp, params=params, metrics=metrics, extra=extra
            ),
            artifacts=artifacts,
        )

    @classmethod
    def from_dict(cls, data: dict[str, t.Any], **kwargs) -> "ModelArtifact":
        try:
            metadata = Metadata.from_dict(data["meta"])
            artifacts = data.get("artifacts")
            serializer = FrameworkSerializers.serializer(metadata.framework.lib)
            model = serializer.load(data["model"].encode(), **kwargs)
        except KeyError:  # pragma: no cover
            raise ValueError("Wrong data")

        if serializer.version() != metadata.framework.version:  # noqa
            warnings.warn(
                f"Model was built using {metadata.framework.lib.value} '{metadata.framework.version}' but detected "
                f"version '{serializer.version()}' installed. This may cause unexpected behavior.",
                exceptions.FrameworkVersionWarning,
            )

        return cls(model=model, meta=metadata, artifacts=artifacts)

    @classmethod
    def from_json(cls, data: str, **kwargs) -> "ModelArtifact":
        return cls.from_dict(json.loads(data), **kwargs)

    @classmethod
    def from_bytes(cls, data: bytes, **kwargs) -> "ModelArtifact":
        return cls.from_json(codecs.decode(data, "zlib"), **kwargs)  # type: ignore[arg-type]

    def to_dict(self, *, artifacts: bool = True, **kwargs) -> dict[str, t.Any]:
        result: dict[str, t.Any] = {
            "model": FrameworkSerializers.serializer(self.meta.framework.lib).dump(self.model, **kwargs).decode(),
            "meta": self.meta.to_dict(),
        }

        if artifacts:
            result["artifacts"] = self.artifacts

        return result

    def to_json(self, *, artifacts: bool = True, **kwargs) -> str:
        return json.dumps(self.to_dict(artifacts=artifacts, **kwargs))

    def to_bytes(self, *, artifacts: bool = True, **kwargs) -> bytes:
        return codecs.encode(self.to_json(artifacts=artifacts, **kwargs).encode(), "zlib")

    def dump(
        self,
        path: t.Union[str, os.PathLike] = "model.flm",
        compression: t.Union[str, Compression] = Compression.standard,
        **kwargs,
    ) -> None:
        """Serialize model artifact into a file.

        :param path: Model file path.
        :param compression: Compression type.
        :param kwargs: Keyword arguments passed to library dump method.
        """
        logger.info("Dump model '%s'", path)
        with tarfile.open(path, f"w:{Compression[compression].compression_format}") as tar:  # type: ignore
            if self.artifacts:
                for name, path in self.artifacts.items():
                    tar.add(path, f"artifacts/{name}")

            with tempfile.NamedTemporaryFile("wb") as model:
                model.write(self.to_bytes(artifacts=False, **kwargs))
                tar.add(model.name, "model")

    @classmethod
    def load(cls, path: t.Union[str, os.PathLike], **kwargs) -> "ModelArtifact":
        """Deserialize model artifact from a file.

        :param path: Model file path.
        :param kwargs: Keyword arguments passed to library load method.
        :return: Model artifact loaded.
        """
        logger.info("Load model '%s'", path)
        model_directory = _ModelDirectory(path)

        with open(model_directory.model, "rb") as f:
            model_artifact = cls.from_bytes(f.read(), **kwargs)

        return cls(
            model=model_artifact.model,
            meta=model_artifact.meta,
            artifacts=model_directory.artifacts or None,
            _directory=model_directory,
        )



import datetime
import typing as t
from unittest.mock import patch

import marshmallow
import pydantic
import pytest
import typesystem

from flama import exceptions
from flama.schemas.validation import output_validation


class TestCaseSchemaValidateOutput:
    @pytest.fixture(scope="function")
    def product_schema(self, app):
        from flama import schemas

        if schemas.lib == pydantic:

            def rating_validator(cls, x):
                assert x >= 0
                return x

            schema = pydantic.create_model(
                "Product",
                name=(str, ...),
                rating=(int, ...),
                created=(t.Optional[datetime.datetime], ...),
                __validators__={"rating": rating_validator},
            )
        elif schemas.lib == typesystem:
            schema = typesystem.Schema(
                title="Product",
                fields={
                    "name": typesystem.fields.String(title="name"),
                    "rating": typesystem.fields.Integer(title="rating", minimum=0),
                    "created": typesystem.fields.DateTime(title="created", allow_null=True),
                },
            )
        elif schemas.lib == marshmallow:
            schema = type(
                "Product",
                (marshmallow.Schema,),
                {
                    "name": marshmallow.fields.String(),
                    "rating": marshmallow.fields.Integer(validate=marshmallow.validate.Range(min=0)),
                    "created": marshmallow.fields.DateTime(allow_none=True),
                },
            )
        else:
            raise ValueError("Wrong schema lib")

        return schema

    @pytest.fixture(scope="function", autouse=True)
    def add_endpoints(self, app, product_schema):
        @app.route("/product", methods=["GET"])
        @output_validation()
        def validate_product() -> product_schema:
            return {
                "name": "foo",
                "rating": 0,
                "created": datetime.datetime(2018, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc),
            }

        @app.route("/many-products", methods=["GET"])
        @output_validation()
        def validate_many_products() -> product_schema:
            return [
                {
                    "name": "foo",
                    "rating": 0,
                    "created": datetime.datetime(2018, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc),
                },
                {
                    "name": "bar",
                    "rating": 1,
                    "created": datetime.datetime(2018, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc),
                },
            ]

        @app.route("/validation-error")
        @output_validation()
        def output_validation_error() -> product_schema:
            return {"name": "foo", "rating": -1}

        class CustomValidationError(exceptions.ValidationError): ...

        @app.route("/custom-error")
        @output_validation(error_cls=CustomValidationError, error_status_code=502)
        def validate_custom_error() -> product_schema:
            return {"name": "foo", "rating": -1}

    @pytest.mark.parametrize(
        ["path", "status_code", "expected_response"],
        [
            pytest.param(
                "/product",
                200,
                {
                    "name": "foo",
                    "rating": 0,
                    "created": "2018-01-01T00:00:00+00:00",
                },
                id="single_item",
            ),
            pytest.param(
                "/many-products",
                200,
                [
                    {
                        "name": "foo",
                        "rating": 0,
                        "created": "2018-01-01T00:00:00+00:00",
                    },
                    {
                        "name": "bar",
                        "rating": 1,
                        "created": "2018-01-01T00:00:00+00:00",
                    },
                ],
                id="many_items",
            ),
            pytest.param(
                "/validation-error",
                500,
                None,
                id="validation_error",
            ),
            pytest.param(
                "/custom-error",
                502,
                None,
                id="custom_error",
            ),
        ],
    )
    async def test_validation(self, client, path, status_code, expected_response):
        response = await client.get(path)
        assert response.status_code == status_code, response.json()
        if status_code == 200:
            assert response.json() == expected_response

    async def test_validation_uncontrolled_error(self, client):
        with patch(
            "flama.schemas.validation.schemas.adapter.dump",
            side_effect=[
                Exception,
                {
                    "detail": "Error serializing response before validation: Exception",
                    "error": "ValidationError",
                    "status_code": 500,
                },
            ],
        ):
            response = await client.get("/product")
            assert response.status_code == 500
            assert response.json() == {
                "detail": "Error serializing response before validation: Exception",
                "error": "ValidationError",
                "status_code": 500,
            }

    def test_function_without_return_schema(self):
        with pytest.raises(TypeError, match=r"Invalid return signature for function .*"):

            @output_validation()
            def foo(): ...



import logging

import flama
from flama import Flama, routing


class AppStatus:
    loaded = False


async def startup():
    logging.info("\nStarting up the ML API...\n")
    # Here, whatever action we want to be run at the startup of the application
    AppStatus.loaded = True


async def shutdown():
    logging.info("\nShutting down the ML API...\n")
    # Here, whatever action we want to be run at the shutdown of the application


def home():
    """
    tags:
        - Home
    summary:
        Returns readiness message
    description:
        The function returns a readiness message in which the global variable AppStatus.loaded is shown.
        If the 'on_startup' function has worked as expected, the message will show the 'loaded' variable as True.
        Else, it'll show the variable as 'False'
    """
    return f"The API is ready. Loaded = {AppStatus.loaded}"


def user_me():
    """
    tags:
        - User
    summary:
        Returns hello 'John Doe'
    description:
        The function returns the plain-text message "Hello John Doe"
    """
    username = "John Doe"
    return f"Hello {username}"


def user(username: str):
    """
    tags:
        - User
    summary:
        Returns hello 'username'
    description:
        The function returns the plain-text message "Hello 'username'" where the 'username' is the user specified as
        query parameter.
    """
    return f"Hello {username}"


app = Flama(
    openapi={
        "info": {
            "title": "Flama ML",
            "version": "0.1.0",
            "description": "Machine learning API using Flama 🔥",
        }
    },
    routes=[
        routing.Route("/", home),
        routing.Route("/user/me", user_me),
        routing.Route("/user/{username}", user),
    ],
    events={"startup": [startup], "shutdown": [shutdown]},
)


app.models.add_model(
    path="/sk_model",
    model="examples/sklearn_model.flm",
    name="logistic-regression",
)


if __name__ == "__main__":
    flama.run(flama_app="__main__:app", server_host="0.0.0.0", server_port=8080, server_reload=True)



import typing as t

import starlette.websockets

if t.TYPE_CHECKING:
    from flama import types

__all__ = ["WebSocket", "Close", "State"]

State = starlette.websockets.WebSocketState


class WebSocket(starlette.websockets.WebSocket):
    def __init__(self, scope: "types.Scope", receive: "types.Receive", send: "types.Send"):  # type: ignore[override]
        super().__init__(scope, receive, send)  # type: ignore[arg-type]

    @property
    def is_connecting(self) -> bool:
        """Check if websocket is connecting.

        :return: True if connecting.
        """
        return self.client_state == starlette.websockets.WebSocketState.CONNECTING

    @property
    def is_connected(self) -> bool:
        """Check if websocket is connected.

        :return: True if connected.
        """
        return self.client_state == starlette.websockets.WebSocketState.CONNECTED

    @property
    def is_disconnected(self) -> bool:
        """Check if websocket is disconnected.

        :return: True if disconnected.
        """
        return self.client_state == starlette.websockets.WebSocketState.DISCONNECTED


class Close(starlette.websockets.WebSocketClose):
    async def __call__(  # type: ignore[override]
        self, scope: "types.Scope", receive: "types.Receive", send: "types.Send"
    ) -> None:
        await super().__call__(scope, receive, send)  # type: ignore[arg-type]



import os
import typing as t

from flama.models.resource import ModelResource, ModelResourceType
from flama.modules import Module

__all__ = ["ModelsModule"]


class ModelsModule(Module):
    name = "models"

    def add_model(
        self,
        path: str,
        model: t.Union[str, os.PathLike],
        name: str,
        tags: t.Optional[dict[str, dict[str, t.Any]]] = None,
        *args,
        **kwargs,
    ) -> ModelResource:
        """Adds a model to this application, setting its endpoints.

        :param path: Resource base path.
        :param model: Model path.
        :param name: Model name.
        :param tags: Tags to add to the model methods.
        """

        name_ = name
        model_ = model

        class Resource(ModelResource, metaclass=ModelResourceType):
            name = name_
            model_path = model_

        resource = Resource()
        self.app.add_component(resource.component)
        self.app.resources.add_resource(path, resource, tags=tags, *args, **kwargs)  # type: ignore[attr-defined]
        return resource

    def model_resource(
        self, path: str, tags: t.Optional[dict[str, dict[str, t.Any]]] = None, *args, **kwargs
    ) -> t.Callable:
        """Decorator for ModelResource classes for adding them to the application.

        :param path: Resource base path.
        :param tags: Tags to add to the model methods.
        :return: Decorated resource class.
        """

        def decorator(resource: type[ModelResource]) -> type[ModelResource]:
            self.app.add_component(resource.component)
            self.app.resources.add_resource(path, resource, tags=tags, *args, **kwargs)  # type: ignore[attr-defined]
            return resource

        return decorator

    def add_model_resource(
        self,
        path: str,
        resource: t.Union[ModelResource, type[ModelResource]],
        tags: t.Optional[dict[str, dict[str, t.Any]]] = None,
        *args,
        **kwargs,
    ) -> ModelResource:
        """Adds a resource to this application, setting its endpoints.

        :param path: Resource base path.
        :param resource: Resource class.
        :param tags: Tags to add to the model methods.
        """
        self.app.add_component(resource.component)
        resource_instance: ModelResource = self.app.resources.add_resource(  # type: ignore[attr-defined]
            path, resource, tags=tags, *args, **kwargs
        )
        return resource_instance



# ruff: noqa
import datetime
import typing as t
import uuid

from marshmallow.fields import *

MAPPING: t.Dict[t.Union[t.Type, None], t.Type[Field]] = {
    None: Field,
    int: Integer,
    float: Float,
    str: String,
    bool: Boolean,
    list: List,
    dict: Dict,
    uuid.UUID: UUID,
    datetime.date: Date,
    datetime.datetime: DateTime,
    datetime.time: Time,
}

MAPPING_TYPES = {v: k for k, v in MAPPING.items()}



import codecs
import importlib.metadata
import json
import tempfile
import typing as t

from flama import exceptions
from flama.serialize import types
from flama.serialize.base import Serializer

try:
    import tensorflow as tf  # type: ignore
except Exception:  # pragma: no cover
    tf = None  # type: ignore[misc, assignment]

if t.TYPE_CHECKING:
    from flama.types import JSONSchema


class TensorFlowSerializer(Serializer):
    lib = types.Framework.tensorflow

    def dump(self, obj: t.Any, **kwargs) -> bytes:
        if tf is None:  # noqa
            raise exceptions.FrameworkNotInstalled("tensorflow")

        with tempfile.NamedTemporaryFile(mode="rb", suffix=".keras") as tmp_file:
            tf.keras.models.save_model(obj, tmp_file.name)  # type: ignore
            return codecs.encode(tmp_file.read(), "base64")

    def load(self, model: bytes, **kwargs) -> t.Any:
        if tf is None:  # noqa
            raise exceptions.FrameworkNotInstalled("tensorflow")

        with tempfile.NamedTemporaryFile(mode="wb", suffix=".keras") as tmp_file:
            tmp_file.write(codecs.decode(model, "base64"))
            return tf.keras.models.load_model(tmp_file.name)  # type: ignore

    def info(self, model: t.Any) -> t.Optional["JSONSchema"]:
        model_info: JSONSchema = json.loads(model.to_json())
        return model_info

    def version(self) -> str:
        for lib in ("tensorflow", "tensorflow-cpu", "tensorflow-gpu", "keras"):
            try:
                return importlib.metadata.version(lib)
            except Exception:
                pass

        raise exceptions.FrameworkNotInstalled("tensorflow")  # noqa



import uuid

import pytest

from flama import Flama, authentication

TOKEN = (
    b"eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9.eyJkYXRhIjogeyJmb28iOiAiYmFyIn0sICJpYXQiOiAwfQ==.J3zdedMZSFNOimstjJat0V"
    b"28rM_b1UU62XCp9dg_5kg="
)


@pytest.fixture(scope="function")
def secret():
    return uuid.UUID(int=0)


@pytest.fixture(scope="function")
def app(secret):
    return Flama(
        schema=None,
        docs=None,
        components=[
            authentication.AccessTokenComponent(secret=secret.bytes),
            authentication.RefreshTokenComponent(secret=secret.bytes),
        ],
    )


class TestCaseAccessTokenComponent:
    @pytest.fixture(scope="function", autouse=True)
    def add_endpoints(self, app):
        @app.get("/")
        def access_token(token: authentication.AccessToken):
            return token.to_dict()

    @pytest.mark.parametrize(
        ["headers", "cookies", "status_code", "result"],
        (
            pytest.param(
                {"access_token": f"Bearer {TOKEN.decode()}"},
                None,
                200,
                {"header": {"alg": "HS256", "typ": "JWT"}, "payload": {"data": {"foo": "bar"}, "iat": 0}},
                id="headers",
            ),
            pytest.param(
                {"access_token": "token"},
                None,
                400,
                {
                    "detail": {
                        "description": "Authentication header must be 'access_token: Bearer <token>'",
                        "error": "JWTException",
                    },
                    "error": "HTTPException",
                    "status_code": 400,
                },
                id="header_wrong_format",
            ),
            pytest.param(
                {"access_token": "Foo token"},
                None,
                400,
                {
                    "detail": {
                        "description": "Authentication header must be 'access_token: Bearer <token>'",
                        "error": "JWTException",
                    },
                    "error": "HTTPException",
                    "status_code": 400,
                },
                id="header_wrong_prefix",
            ),
            pytest.param(
                None,
                {"access_token": TOKEN.decode()},
                200,
                {"header": {"alg": "HS256", "typ": "JWT"}, "payload": {"data": {"foo": "bar"}, "iat": 0}},
                id="cookies",
            ),
            pytest.param(
                None,
                None,
                401,
                {"detail": "Unauthorized", "error": "HTTPException", "status_code": 401},
                id="unauthorized",
            ),
            pytest.param(
                None,
                {
                    "access_token": "eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9.eyJkYXRhIjogeyJmb28iOiAiYmFyI"
                    "n0sICJpYXQiOiAwfQ==.0000",
                },
                401,
                {
                    "detail": {
                        "description": "Signature verification failed for token 'eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVC"
                        "J9.eyJkYXRhIjogeyJmb28iOiAiYmFyIn0sICJpYXQiOiAwfQ==.0000'",
                        "error": "JWTValidateException",
                    },
                    "error": "HTTPException",
                    "status_code": 401,
                },
                id="invalid_token",
            ),
        ),
    )
    async def test_injection(self, client, status_code, result, headers, cookies):
        client.headers = headers
        client.cookies = cookies
        response = await client.request("get", "/")

        assert response.status_code == status_code
        assert response.json() == result


class TestCaseRefreshTokenComponent:
    @pytest.fixture(scope="function", autouse=True)
    def add_endpoints(self, app):
        @app.get("/")
        def refresh_token(token: authentication.RefreshToken):
            return token.to_dict()

    @pytest.mark.parametrize(
        ["headers", "cookies", "status_code", "result"],
        (
            pytest.param(
                {"refresh_token": f"Bearer {TOKEN.decode()}"},
                None,
                200,
                {"header": {"alg": "HS256", "typ": "JWT"}, "payload": {"data": {"foo": "bar"}, "iat": 0}},
                id="headers",
            ),
            pytest.param(
                {"refresh_token": "token"},
                None,
                400,
                {
                    "detail": {
                        "description": "Authentication header must be 'refresh_token: Bearer <token>'",
                        "error": "JWTException",
                    },
                    "error": "HTTPException",
                    "status_code": 400,
                },
                id="header_wrong_format",
            ),
            pytest.param(
                {"refresh_token": "Foo token"},
                None,
                400,
                {
                    "detail": {
                        "description": "Authentication header must be 'refresh_token: Bearer <token>'",
                        "error": "JWTException",
                    },
                    "error": "HTTPException",
                    "status_code": 400,
                },
                id="header_wrong_prefix",
            ),
            pytest.param(
                None,
                {"refresh_token": TOKEN.decode()},
                200,
                {"header": {"alg": "HS256", "typ": "JWT"}, "payload": {"data": {"foo": "bar"}, "iat": 0}},
                id="cookies",
            ),
            pytest.param(
                None,
                None,
                401,
                {"detail": "Unauthorized", "error": "HTTPException", "status_code": 401},
                id="unauthorized",
            ),
            pytest.param(
                None,
                {
                    "refresh_token": "eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9.eyJkYXRhIjogeyJmb28iOiAiYmFyI"
                    "n0sICJpYXQiOiAwfQ==.0000",
                },
                401,
                {
                    "detail": {
                        "description": "Signature verification failed for token 'eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVC"
                        "J9.eyJkYXRhIjogeyJmb28iOiAiYmFyIn0sICJpYXQiOiAwfQ==.0000'",
                        "error": "JWTValidateException",
                    },
                    "error": "HTTPException",
                    "status_code": 401,
                },
                id="invalid_token",
            ),
        ),
    )
    async def test_injection(self, client, status_code, result, headers, cookies):
        client.headers = headers
        client.cookies = cookies
        response = await client.request("get", "/")

        assert response.status_code == status_code
        assert response.json() == result



from unittest.mock import AsyncMock, Mock, call, patch

import pytest
import sqlalchemy
from sqlalchemy.ext.asyncio import AsyncConnection, AsyncEngine, AsyncTransaction, create_async_engine

from flama import Flama
from flama.client import Client
from flama.exceptions import ApplicationError, SQLAlchemyError
from flama.sqlalchemy import ConnectionManager, MultipleConnectionManager, SingleConnectionManager, SQLAlchemyModule


@pytest.fixture(scope="function")
async def app():
    return Flama(modules={SQLAlchemyModule("sqlite+aiosqlite://")})


@pytest.fixture(scope="function")
def engine():
    return create_async_engine("sqlite+aiosqlite://")


class TestCaseSingleConnectionManager:
    @pytest.fixture(scope="function")
    def connection_manager(self, engine):
        return SingleConnectionManager(engine)

    @pytest.fixture(scope="function")
    async def connection(self, connection_manager):
        c = await connection_manager.open()

        yield c

        await connection_manager.close(c)

    def test_init(self, engine):
        connection_manager = SingleConnectionManager(engine)

        assert connection_manager._connection is None
        assert connection_manager._transaction is None
        assert connection_manager._connection_clients == 0

    def test_connection_not_initialized(self, connection_manager):
        with pytest.raises(SQLAlchemyError, match="Connection not initialized"):
            connection_manager.connection

    def test_transaction_not_initialized(self, connection_manager):
        with pytest.raises(SQLAlchemyError, match="Transaction not started"):
            connection_manager.transaction

    async def test_open_close_connection(self, connection_manager):
        assert connection_manager._connection_clients == 0
        assert connection_manager._connection is None

        first_connection = await connection_manager.open()

        assert connection_manager._connection_clients == 1
        assert connection_manager._connection == first_connection

        second_connection = await connection_manager.open()

        assert first_connection == second_connection
        assert connection_manager._connection_clients == 2
        assert connection_manager._connection == second_connection

        # Close a wrong connection
        with pytest.raises(SQLAlchemyError, match="Wrong connection"):
            await connection_manager.close(Mock(spec=AsyncConnection))

        await connection_manager.close(second_connection)

        assert connection_manager._connection_clients == 1
        assert connection_manager._connection == first_connection

        await connection_manager.close(first_connection)

        assert connection_manager._connection_clients == 0
        assert connection_manager._connection is None

    @pytest.mark.parametrize(
        ["rollback", "commit_calls", "rollback_calls"],
        (
            pytest.param(False, [call()], [], id="commit"),
            pytest.param(True, [], [call()], id="rollback"),
        ),
    )
    async def test_begin_end_transaction(self, connection_manager, connection, rollback, commit_calls, rollback_calls):
        # Begin a transaction using a wrong connection
        with pytest.raises(SQLAlchemyError, match="Wrong connection"):
            await connection_manager.begin(Mock(spec=AsyncConnection))

        assert connection_manager._transaction_clients == 0
        assert connection_manager._transaction is None

        with patch.multiple("sqlalchemy.ext.asyncio.AsyncTransaction", rollback=AsyncMock(), commit=AsyncMock()):
            first_transaction = await connection_manager.begin(connection)

            assert connection_manager._transaction_clients == 1
            assert connection_manager._transaction == first_transaction

            second_transaction = await connection_manager.begin(connection)

            assert first_transaction == second_transaction
            assert connection_manager._transaction_clients == 2
            assert connection_manager._transaction == first_transaction

            # Close a transaction using a wrong transaction
            with pytest.raises(SQLAlchemyError, match="Wrong transaction"):
                await connection_manager.end(Mock(spec=AsyncTransaction))

            await connection_manager.end(second_transaction, rollback=rollback)

            assert second_transaction.commit.call_args_list == []
            assert second_transaction.rollback.call_args_list == []
            assert connection_manager._transaction_clients == 1
            assert connection_manager._transaction == first_transaction

            await connection_manager.end(first_transaction, rollback=rollback)

            assert first_transaction.commit.call_args_list == commit_calls
            assert first_transaction.rollback.call_args_list == rollback_calls
            assert connection_manager._transaction_clients == 0
            assert connection_manager._transaction is None


class TestCaseMultipleConnectionManager:
    @pytest.fixture(scope="function")
    def connection_manager(self, engine):
        return MultipleConnectionManager(engine)

    def test_init(self, engine):
        connection_manager = MultipleConnectionManager(engine)

        assert connection_manager._connections == set()
        assert connection_manager._transactions == {}

    async def test_open_close_connection(self, connection_manager):
        assert connection_manager._connections == set()

        first_connection = await connection_manager.open()

        assert connection_manager._connections == {first_connection}

        second_connection = await connection_manager.open()

        assert first_connection != second_connection
        assert connection_manager._connections == {first_connection, second_connection}

        # Close an uninitialized connection
        with pytest.raises(SQLAlchemyError, match="Connection not initialized"):
            await connection_manager.close(Mock(spec=AsyncConnection))

        await connection_manager.close(second_connection)

        assert connection_manager._connections == {first_connection}

        await connection_manager.close(first_connection)

        assert connection_manager._connections == set()

    @pytest.mark.parametrize(
        ["rollback", "commit_calls", "rollback_calls"],
        (
            pytest.param(False, [call()], [], id="commit"),
            pytest.param(True, [], [call()], id="rollback"),
        ),
    )
    async def test_begin_end_transaction(self, connection_manager, rollback, commit_calls, rollback_calls):
        assert connection_manager._connections == set()
        assert connection_manager._transactions == {}

        # Begin a transaction using a wrong connection
        with pytest.raises(SQLAlchemyError, match="Connection not initialized"):
            await connection_manager.begin(Mock(spec=AsyncConnection))

        # Close a transaction that is not started
        with pytest.raises(SQLAlchemyError, match="Transaction not started"):
            await connection_manager.end(Mock(spec=AsyncTransaction))

        with patch.multiple("sqlalchemy.ext.asyncio.AsyncTransaction", rollback=AsyncMock(), commit=AsyncMock()):
            first_connection = await connection_manager.open()
            second_connection = await connection_manager.open()
            first_connection_transaction = await connection_manager.begin(first_connection)
            second_connection_transaction = await connection_manager.begin(second_connection)

            # Two transactions for the same connection
            with pytest.raises(SQLAlchemyError, match="Transaction already started in this connection"):
                await connection_manager.begin(first_connection)

            assert connection_manager._connections == {first_connection, second_connection}
            assert connection_manager._transactions == {
                first_connection: first_connection_transaction,
                second_connection: second_connection_transaction,
            }

            await connection_manager.end(second_connection_transaction, rollback=rollback)
            assert second_connection_transaction.commit.call_args_list == commit_calls
            assert second_connection_transaction.rollback.call_args_list == rollback_calls

            assert connection_manager._transactions == {first_connection: first_connection_transaction}

            await connection_manager.close(first_connection)

            assert connection_manager._connections == {second_connection}
            assert connection_manager._transactions == {}

            await connection_manager.close(second_connection)

            assert connection_manager._connections == set()
            assert connection_manager._transactions == {}


class TestCaseSQLAlchemyModule:
    @pytest.mark.parametrize(
        ["uri", "exception"],
        (
            pytest.param("sqlite+aiosqlite://", None, id="sqlite"),
            pytest.param(None, ApplicationError("Database connection string must be provided"), id="no_database"),
        ),
        indirect=["exception"],
    )
    async def test_lifespan_cycle(self, uri, exception):
        with exception:
            app = Flama(modules={SQLAlchemyModule(uri)})
            async with Client(app):  # Initialize app life-cycle
                assert app.sqlalchemy.database == uri
                assert isinstance(app.sqlalchemy.metadata, sqlalchemy.MetaData)
                assert isinstance(app.sqlalchemy._engine, AsyncEngine)
                assert isinstance(app.sqlalchemy._connection_manager, ConnectionManager)

            assert app.sqlalchemy.database == uri
            assert isinstance(app.sqlalchemy.metadata, sqlalchemy.MetaData)
            assert app.sqlalchemy._engine is None
            assert app.sqlalchemy._connection_manager is None

    def test_engine_not_initialized(self, app):
        with pytest.raises(ApplicationError, match="SQLAlchemyModule not initialized"):
            app.sqlalchemy.engine

    def test_connection_manager_not_initialized(self, app):
        with pytest.raises(ApplicationError, match="SQLAlchemyModule not initialized"):
            app.sqlalchemy.connection_manager

    async def test_open_connection(self, app):
        connection_manager_mock = AsyncMock(spec=ConnectionManager)

        with patch.object(app.sqlalchemy, "_connection_manager", connection_manager_mock):
            await app.sqlalchemy.open_connection()

        assert connection_manager_mock.open.await_args_list == [call()]

    async def test_close_connection(self, app):
        connection_manager_mock = AsyncMock(spec=ConnectionManager)
        connection_mock = Mock(spec=AsyncConnection)

        with patch.object(app.sqlalchemy, "_connection_manager", connection_manager_mock):
            await app.sqlalchemy.close_connection(connection_mock)

        assert connection_manager_mock.close.await_args_list == [call(connection_mock)]

    async def test_begin_transaction(self, app):
        connection_manager_mock = AsyncMock(spec=ConnectionManager)
        connection_mock = Mock(spec=AsyncConnection)

        with patch.object(app.sqlalchemy, "_connection_manager", connection_manager_mock):
            await app.sqlalchemy.begin_transaction(connection_mock)

        assert connection_manager_mock.begin.await_args_list == [call(connection_mock)]

    async def test_end_transaction(self, app):
        connection_manager_mock = AsyncMock(spec=ConnectionManager)
        transaction_mock = Mock(spec=AsyncTransaction)

        with patch.object(app.sqlalchemy, "_connection_manager", connection_manager_mock):
            await app.sqlalchemy.end_transaction(transaction_mock)

        assert connection_manager_mock.end.await_args_list == [call(transaction_mock, rollback=False)]



import typing as t

from flama import exceptions
from flama.codecs.base import HTTPCodec

if t.TYPE_CHECKING:
    from flama import http

__all__ = ["JSONDataCodec"]


class JSONDataCodec(HTTPCodec):
    media_type = "application/json"
    format = "json"

    async def decode(self, item: "http.Request", **options):
        try:
            if await item.body() == b"":
                return None

            return await item.json()
        except ValueError as exc:
            raise exceptions.DecodeError(f"Malformed JSON. {exc}") from None



import marshmallow
from marshmallow import Schema
from marshmallow.fields import Field

from flama.schemas._libs.marshmallow import fields, schemas
from flama.schemas._libs.marshmallow.adapter import MarshmallowAdapter

lib = marshmallow
adapter = MarshmallowAdapter()

__all__ = ["Field", "Schema", "fields", "adapter", "lib", "schemas"]



import flama
from flama import Flama, routing

app = Flama(
    openapi={
        "info": {
            "title": "Hello-🔥",
            "version": "1.0",
            "description": "My first API",
        },
        "tags": [
            {"name": "Salute", "description": "This is the salute description"},
        ],
    },
    debug=True,
)


@app.route("/")
def home():
    """
    tags:
        - Salute
    summary:
        Returns a warming message.
    description:
        This is a more detailed description of the method itself.
        Here we can give all the details required and they will appear
        automatically in the auto-generated docs.
    responses:
        200:
            description: Warming hello message!
    """
    return {"message": "Hello 🔥"}


error_app = routing.Router()


class FooException(Exception): ...


@error_app.route("/500/")
def error_500():
    """
    tags:
        - Error
    summary:
        Returns an Internal Server Error.
    description:
        Returns an Internal Server Error describing the error occurred and the whole trace.
    responses:
        500:
            description: Internal Server Error
    """
    raise FooException("Foo")


app.mount("/error", app=error_app)

# Mount a complex urls tree to illustrate in 404 error page
bar_app = routing.Router()


@bar_app.route("/foobar/")
def foobar(): ...


@bar_app.route("/barfoo/")
def barfoo(): ...


foo_app = routing.Router()
foo_app.mount("/bar/", bar_app)

app.mount("/foo/", foo_app)

if __name__ == "__main__":
    flama.run(flama_app=app, server_host="0.0.0.0", server_port=8080)
